# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import Schema, fields, post_dump, post_load

from polyaxon_schemas.base import BaseConfig, BaseMultiSchema


class BaseRegularizerSchema(Schema):
    name = fields.Str(allow_none=True)
    collect = fields.Bool(default=True, missing=True)


class BaseRegularizerConfig(BaseConfig):
    REDUCED_ATTRIBUTES = ['name']

    def __init__(self, name, collect=True):
        self.name = name
        self.collect = collect


class L1RegularizerSchema(BaseRegularizerSchema):
    l = fields.Float(default=0.01, missing=0.01)  # noqa

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return L1RegularizerConfig(**data)

    @post_dump
    def unmake(self, data):
        return L1RegularizerConfig.remove_reduced_attrs(data)


class L1RegularizerConfig(BaseRegularizerConfig):
    """Regularizer for L1 regularization.

    Args:
        l: Float; regularization factor.

    Polyaxonfile usage:

    ```yaml
    Dense:
      units: 32
      kernel_regularizer: L1
    ```

    or

    ```yaml
    Dense:
      units: 32
      kernel_regularizer:
        L1:
          l: 0.2
    ```

    or

    ```yaml
    Dense:
      units: 32
      kernel_regularizer:
        L1: {l: 0.2}
    ```
    """
    IDENTIFIER = 'L1'
    SCHEMA = L1RegularizerSchema

    def __init__(self, l=0.01, name='L1Regularizer', collect=True):  # noqa
        self.l = l  # noqa
        super(L1RegularizerConfig, self).__init__(name, collect)


class L2RegularizerSchema(BaseRegularizerSchema):
    l = fields.Float(default=0.01, missing=0.01)  # noqa

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return L2RegularizerConfig(**data)

    @post_dump
    def unmake(self, data):
        return L2RegularizerConfig.remove_reduced_attrs(data)


class L2RegularizerConfig(BaseRegularizerConfig):
    """Regularizer for L2 regularization.

    Args:
        l: Float; regularization factor.

    Polyaxonfile usage:

    ```yaml
    Dense:
      units: 32
      kernel_regularizer: L2
    ```

    or

    ```yaml
    Dense:
      units: 32
      kernel_regularizer:
        L2:
          l: 0.2
    ```

    or

    ```yaml
    Dense:
      units: 32
      kernel_regularizer:
        L2: {l: 0.2}
    ```
    """
    IDENTIFIER = 'L2'
    SCHEMA = L2RegularizerSchema

    def __init__(self, l=0.01, name='L2Regularizer', collect=True):  # noqa
        self.l = l  # noqa
        super(L2RegularizerConfig, self).__init__(name, collect)


class L1L2RegularizerSchema(BaseRegularizerSchema):
    l1 = fields.Float(default=0.01, missing=0.01)
    l2 = fields.Float(default=0.01, missing=0.01)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return L1L2RegularizerConfig(**data)

    @post_dump
    def unmake(self, data):
        return L1L2RegularizerConfig.remove_reduced_attrs(data)


class L1L2RegularizerConfig(BaseRegularizerConfig):
    """Regularizer for L1 and L2 regularization.

    Args:
        l1: Float; L1 regularization factor.
        l2: Float; L2 regularization factor.

    Polyaxonfile usage:

    ```yaml
    Dense:
      units: 32
      kernel_regularizer: L1L2
    ```

    or

    ```yaml
    Dense:
      units: 32
      kernel_regularizer:
        L1L2:
          l1: 0.2
          l2: 0.1
    ```

    or

    ```yaml
    Dense:
      units: 32
      kernel_regularizer:
        L1L2: {l1: 0.2, l2: 0.1}
    ```
    """
    IDENTIFIER = 'L1L2'
    SCHEMA = L1L2RegularizerSchema

    def __init__(self, l1=0.01, l2=0.01, name='L1L2Regularizer', collect=True):
        self.l1 = l1
        self.l2 = l2
        super(L1L2RegularizerConfig, self).__init__(name, collect)


class RegularizerSchema(BaseMultiSchema):
    __multi_schema_name__ = 'regularizer'
    __configs__ = {
        L1RegularizerConfig.IDENTIFIER: L1RegularizerConfig,
        L2RegularizerConfig.IDENTIFIER: L2RegularizerConfig,
        L1L2RegularizerConfig.IDENTIFIER: L1L2RegularizerConfig,
    }
    __support_snake_case__ = True



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, post_dump, post_load

from polyaxon_schemas.ml.constraints import ConstraintSchema
from polyaxon_schemas.ml.initializations import InitializerSchema, UniformInitializerConfig
from polyaxon_schemas.ml.layers.base import BaseLayerConfig, BaseLayerSchema
from polyaxon_schemas.ml.regularizations import RegularizerSchema


class EmbeddingSchema(BaseLayerSchema):
    input_dim = fields.Int()
    output_dim = fields.Int()
    embeddings_initializer = fields.Nested(InitializerSchema, allow_none=True)
    embeddings_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    activity_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    embeddings_constraint = fields.Nested(ConstraintSchema, allow_none=True)
    mask_zero = fields.Bool(default=False, missing=False)
    input_length = fields.Int(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return EmbeddingConfig(**data)

    @post_dump
    def unmake(self, data):
        return EmbeddingConfig.remove_reduced_attrs(data)


class EmbeddingConfig(BaseLayerConfig):
    """Turns positive integers (indexes) into dense vectors of fixed size.

    eg. [[4], [20]] -> [[0.25, 0.1], [0.6, -0.2]]

    This layer can only be used as the first layer in a model.

    Args:
        input_dim: int > 0. Size of the vocabulary,
            i.e. maximum integer index + 1.
        output_dim: int >= 0. Dimension of the dense embedding.
        embeddings_initializer: Initializer for the `embeddings` matrix.
        embeddings_regularizer: Regularizer function applied to
              the `embeddings` matrix.
        embeddings_constraint: Constraint function applied to
              the `embeddings` matrix.
        mask_zero: Whether or not the input value 0 is a special "padding"
            value that should be masked out.
            This is useful when using recurrent layers,
            which may take variable length inputs.
            If this is `True` then all subsequent layers
            in the model need to support masking or an exception will be raised.
            If mask_zero is set to True, as a consequence, index 0 cannot be
            used in the vocabulary (input_dim should equal size of
            vocabulary + 1).
        input_length: Length of input sequences, when it is constant.
            This argument is required if you are going to connect
            `Flatten` then `Dense` layers upstream
            (without it, the shape of the dense outputs cannot be computed).

    Input shape:
        2D tensor with shape: `(batch_size, sequence_length)`.

    Output shape:
        3D tensor with shape: `(batch_size, sequence_length, output_dim)`.

    References:
        - [A Theoretically Grounded Application of Dropout in Recurrent Neural
          Networks](http://arxiv.org/abs/1512.05287)

    Example:

    ```python
    # input_length=10
    x = Embedding(1000, 64)(x)
    # the model will take as input an integer matrix of size (batch, input_length).
    # the largest integer (i.e. word index) in the input should be no larger
    than 999 (vocabulary size).
    # now x.output_shape == (None, 10, 64), where None is the batch
    dimension.
    ```

    Polyaxonfile usage:

    ```yaml
    Embedding:
      input_dim: 1000
      output_dim: 32
    ```
    """
    IDENTIFIER = 'Embedding'
    SCHEMA = EmbeddingSchema

    def __init__(self,
                 input_dim,
                 output_dim,
                 embeddings_initializer=UniformInitializerConfig(),
                 embeddings_regularizer=None,
                 activity_regularizer=None,
                 embeddings_constraint=None,
                 mask_zero=False,
                 input_length=None,
                 **kwargs):
        super(EmbeddingConfig, self).__init__(**kwargs)
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.embeddings_initializer = embeddings_initializer
        self.embeddings_regularizer = embeddings_regularizer
        self.activity_regularizer = activity_regularizer
        self.embeddings_constraint = embeddings_constraint
        self.mask_zero = mask_zero
        self.input_length = input_length



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import six

from collections import Mapping


def assert_equal_dict(dict1, dict2):
    for k, v in six.iteritems(dict1):
        if v is None:
            continue
        if isinstance(v, Mapping):
            assert_equal_dict(v, dict2[k])
        else:
            assert v == dict2[k]


def assert_equal_feature_processors(fp1, fp2):
    # Check that they have same features
    assert list(six.iterkeys(fp1)) == list(six.iterkeys(fp2))

    # Check that all features have the same graph
    for feature in fp1:
        assert_equal_graphs(fp2[feature], fp1[feature])


def assert_tensors(tensor1, tensor2):
    if isinstance(tensor1, six.string_types):
        tensor1 = [tensor1, 0, 0]

    if isinstance(tensor2, six.string_types):
        tensor2 = [tensor2, 0, 0]

    assert tensor1 == tensor2


def assert_equal_graphs(result_graph, expected_graph):
    for i, input_layer in enumerate(expected_graph['input_layers']):
        assert_tensors(input_layer, result_graph['input_layers'][i])

    for i, output_layer in enumerate(expected_graph['output_layers']):
        assert_tensors(output_layer, result_graph['output_layers'][i])

    for layer_i, layer in enumerate(result_graph['layers']):
        layer_name, layer_data = list(six.iteritems(layer))[0]
        assert layer_name in expected_graph['layers'][layer_i]
        for k, v in six.iteritems(layer_data):
            print(k)
            print('v == {}, new_v: {}'.format(v, expected_graph['layers'][layer_i][layer_name][k]))
            assert v == expected_graph['layers'][layer_i][layer_name][k]


def assert_equal_layers(config, expected_layer):
    result_layer = config.to_dict()
    for k, v in six.iteritems(expected_layer):
        if v is not None or k not in config.REDUCED_ATTRIBUTES:
            assert v == result_layer[k]
        else:
            assert k not in result_layer



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from polyaxon_schemas.user import UserConfig


class TestUserConfigs(TestCase):
    def test_user_config(self):
        config_dict = {'username': 'username',
                       'email': 'user@domain.com',
                       'is_superuser': False}
        config = UserConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import uuid

from unittest import TestCase

from hestia.tz_utils import local_now

from polyaxon_schemas.experiment import (
    ContainerGPUResourcesConfig,
    ContainerResourcesConfig,
    ExperimentConfig,
    ExperimentJobConfig,
    ExperimentJobStatusConfig,
    ExperimentMetricConfig,
    ExperimentStatusConfig
)


class TestExperimentConfigs(TestCase):
    def test_experiment_config(self):
        config_dict = {
            'uuid': uuid.uuid4().hex,
            'project': 'name.name',
            'experiment_group': 'name.name.1',
            'build_job': 'name.name.1',
            'unique_name': 'user.proj.1',
            'last_status': 'Running',
            'description': 'description',
            'config': {'k': 'v'},
            'tags': ['tag1'],
            'num_jobs': 1,
            'created_at': local_now().isoformat(),
            'updated_at': local_now().isoformat(),
            'has_tensorboard': True,
        }
        config = ExperimentConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        config_to_dict.pop('declarations')
        config_to_dict.pop('finished_at')
        config_to_dict.pop('is_clone')
        config_to_dict.pop('jobs')
        config_to_dict.pop('last_metric')
        config_to_dict.pop('resources')
        config_to_dict.pop('run_env')
        config_to_dict.pop('ttl')
        config_to_dict.pop('id')
        config_to_dict.pop('started_at')
        config_to_dict.pop('total_run')
        config_to_dict.pop('user')
        config_to_dict.pop('name')
        assert config_to_dict == config_dict

        config_to_dict = config.to_light_dict()
        config_dict.pop('uuid')
        config_dict.pop('description')
        config_dict.pop('config')
        config_dict.pop('project')
        config_dict.pop('updated_at')
        config_dict.pop('tags')
        config_dict.pop('num_jobs')
        config_to_dict.pop('finished_at')
        config_to_dict.pop('id')
        config_to_dict.pop('started_at')
        config_dict.pop('has_tensorboard')
        config_to_dict.pop('total_run')
        config_to_dict.pop('user')
        assert config_to_dict == config_dict

        config_to_dict = config.to_light_dict(humanize_values=True)
        assert config_to_dict.pop('created_at') == 'a few seconds ago'
        assert config_to_dict.pop('started_at') is None

    def test_experiment_with_jobs_config(self):
        config_dict = {'id': 2,
                       'config': {},
                       'unique_name': 'adam.proj.1',
                       'uuid': uuid.uuid4().hex,
                       'project': 'user.name',
                       'experiment_group': 'user.name.1',
                       'last_status': 'Running',
                       'num_jobs': 1,
                       'created_at': local_now().isoformat(),
                       'updated_at': local_now().isoformat(),
                       'started_at': local_now().isoformat(),
                       'finished_at': local_now().isoformat(),
                       'has_tensorboard': False,
                       'tags': ['tag'],
                       'jobs': [ExperimentJobConfig(uuid=uuid.uuid4().hex,
                                                    experiment=2,
                                                    created_at=local_now(),
                                                    updated_at=local_now(),
                                                    definition={}).to_dict()]}
        config = ExperimentConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert config_to_dict.pop('total_run') == '0s'
        config_to_dict.pop('declarations')
        config_to_dict.pop('description')
        config_to_dict.pop('is_clone')
        config_to_dict.pop('last_metric')
        config_to_dict.pop('resources')
        config_to_dict.pop('run_env')
        config_to_dict.pop('user')
        config_to_dict.pop('name')
        config_to_dict.pop('build_job')
        config_to_dict.pop('ttl')
        assert config_to_dict == config_dict

        config_dict.pop('tags')
        config_to_dict = config.to_light_dict(humanize_values=True)
        assert config_to_dict.pop('total_run') == '0s'
        assert config_to_dict.pop('created_at') == 'a few seconds ago'
        assert config_to_dict.pop('started_at') == 'a few seconds ago'
        assert config_to_dict.pop('finished_at') == 'a few seconds ago'

    def test_experiment_job_config(self):
        config_dict = {'uuid': uuid.uuid4().hex,
                       'experiment': 1,
                       'created_at': local_now().isoformat(),
                       'updated_at': local_now().isoformat(),
                       'started_at': local_now().isoformat(),
                       'finished_at': local_now().isoformat(),
                       'definition': {},
                       'role': 'master',
                       'id': 1,
                       'pod_id': 'job_1',
                       'unique_name': 'project.1.1.master'}
        config = ExperimentJobConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert config_to_dict.pop('total_run') == '0s'
        config_to_dict.pop('last_status')
        config_to_dict.pop('resources')
        assert config_to_dict == config_dict

        config_dict.pop('definition')
        config_dict.pop('experiment')
        config_dict.pop('updated_at')
        config_dict.pop('uuid')
        config_to_dict = config.to_light_dict()
        assert config_to_dict.pop('total_run') == '0s'
        config_dict.pop('unique_name')
        config_to_dict.pop('last_status')
        assert config_to_dict == config_dict

        config_to_dict = config.to_light_dict(humanize_values=True)
        assert config_to_dict.pop('total_run') == '0s'
        assert config_to_dict.pop('created_at') == 'a few seconds ago'
        assert config_to_dict.pop('started_at') == 'a few seconds ago'
        assert config_to_dict.pop('finished_at') == 'a few seconds ago'

    def test_experiment_status_config(self):
        config_dict = {'id': 1,
                       'uuid': uuid.uuid4().hex,
                       'experiment': 1,
                       'created_at': local_now().isoformat(),
                       'status': 'Running',
                       'message': None}
        config = ExperimentStatusConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert config_to_dict == config_dict

        config_dict.pop('experiment', None)
        config_dict.pop('uuid', None)
        config_to_dict = config.to_light_dict()
        assert config_to_dict == config_dict

        config_to_dict = config.to_dict(humanize_values=True)
        assert config_to_dict.pop('created_at') == 'a few seconds ago'

    def test_experiment_metric_config(self):
        config_dict = {'id': 1,
                       'uuid': uuid.uuid4().hex,
                       'experiment': 1,
                       'created_at': local_now().isoformat(),
                       'values': {'accuracy': 0.9}}
        config = ExperimentMetricConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert config_to_dict == config_dict

        config_dict.pop('experiment', None)
        config_dict.pop('uuid', None)
        config_to_dict = config.to_light_dict()
        assert config_to_dict == config_dict

        config_to_dict = config.to_dict(humanize_values=True)
        assert config_to_dict.pop('created_at') == 'a few seconds ago'

    def test_experiment_job_status_config(self):
        config_dict = {'id': 1,
                       'uuid': uuid.uuid4().hex,
                       'job': 1,
                       'created_at': local_now().isoformat(),
                       'status': 'Running'}
        config = ExperimentJobStatusConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        config_to_dict.pop('details')
        config_to_dict.pop('message')
        assert config_to_dict == config_dict

        config_to_dict = config.to_light_dict()
        config_dict.pop('details', None)
        config_dict.pop('job', None)
        config_dict.pop('uuid', None)
        config_to_dict.pop('message')
        assert config_to_dict == config_dict

        config_to_dict = config.to_light_dict(humanize_values=True)
        assert config_to_dict.pop('created_at') == 'a few seconds ago'

    def test_container_gpu_resources(self):
        config_dict = {
            'index': 0,
            'bus_id': '0000:00:1E.1',
            'memory_free': 10000,
            'memory_total': 12883853312,
            'memory_used': 8388608000,
            'memory_utilization': 0,
            'minor': 1,
            'name': 'GeForce GTX TITAN 0',
            'power_draw': 125,
            'power_limit': 250,
            'processes': [{'command': 'python',
                           'gpu_memory_usage': 4000,
                           'pid': 48448,
                           'username': 'user1'},
                          {'command': 'python',
                           'gpu_memory_usage': 4000,
                           'pid': 153223,
                           'username': 'user2'}],
            'serial': '0322917092147',
            'temperature_gpu': 80,
            'utilization_gpu': 76,
            'uuid': 'GPU-10fb0fbd-2696-43f3-467f-d280d906a107'
        }
        config = ContainerGPUResourcesConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert config_to_dict == config_dict

        config_to_dict = config.to_dict(humanize_values=True)
        assert config_to_dict.pop('memory_free') == '9.77 Kb'
        assert config_to_dict.pop('memory_total') == '12.0 Gb'
        assert config_to_dict.pop('memory_used') == '7.81 Gb'

    def test_container_resources(self):
        gpu_resources = {
            'index': 0,
            'bus_id': '0000:00:1E.1',
            'memory_free': 10000,
            'memory_total': 12883853312,
            'memory_used': 8388608000,
            'memory_utilization': 0,
            'minor': 1,
            'name': 'GeForce GTX TITAN 0',
            'power_draw': 125,
            'power_limit': 250,
            'processes': [{'command': 'python',
                           'gpu_memory_usage': 4000,
                           'pid': 48448,
                           'username': 'user1'},
                          {'command': 'python',
                           'gpu_memory_usage': 4000,
                           'pid': 153223,
                           'username': 'user2'}],
            'serial': '0322917092147',
            'temperature_gpu': 80,
            'utilization_gpu': 76,
            'uuid': 'GPU-10fb0fbd-2696-43f3-467f-d280d906a107'
        }

        config_dict = {
            'job_uuid': uuid.uuid4().hex,
            'experiment_uuid': uuid.uuid4().hex,
            'job_name': 'master.0',
            'container_id': '3175e88873af9077688cee20eaadc0c07746efb84d01ae696d6d17ed9bcdfbc4',
            'n_cpus': 2,
            'cpu_percentage': 0.6947691836734693,
            'percpu_percentage': [0.4564075715616173, 0.23836161211185192],
            'memory_used': 84467712,
            'memory_limit': 2096160768,
            'gpu_resources': [gpu_resources]
        }
        config = ContainerResourcesConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert config_to_dict == config_dict

        config_to_dict = config.to_dict(humanize_values=True)
        assert config_to_dict.pop('cpu_percentage') == '69.48%'
        assert config_to_dict.pop('memory_limit') == '1.95 Gb'
        assert config_to_dict.pop('memory_used') == '80.55 Mb'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, post_dump, post_load, validate

from polyaxon_schemas.ml.constraints import ConstraintSchema
from polyaxon_schemas.ml.initializations import (
    GlorotNormalInitializerConfig,
    InitializerSchema,
    ZerosInitializerConfig
)
from polyaxon_schemas.ml.layers.base import BaseLayerConfig, BaseLayerSchema
from polyaxon_schemas.ml.regularizations import RegularizerSchema
from polyaxon_schemas.utils import ACTIVATION_VALUES, DType, StrOrFct


class MaskingSchema(BaseLayerSchema):
    mask_value = fields.Int()

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return MaskingConfig(**data)

    @post_dump
    def unmake(self, data):
        return MaskingConfig.remove_reduced_attrs(data)


class MaskingConfig(BaseLayerConfig):
    """Masks a sequence by using a mask value to skip timesteps.

    For each timestep in the input tensor (dimension #1 in the tensor),
    if all values in the input tensor at that timestep
    are equal to `mask_value`, then the timestep will be masked (skipped)
    in all downstream layers (as long as they support masking).

    If any downstream layer does not support masking yet receives such
    an input mask, an exception will be raised.

    Example:

    Consider a Numpy data array `x` of shape `(samples, timesteps, features)`,
    to be fed to a LSTM layer.
    You want to mask timestep #3 and #5 because you lack data for
    these timesteps. You can:

        - set `x[:, 3, :] = 0.` and `x[:, 5, :] = 0.`
        - insert a `Masking` layer with `mask_value=0.` before the LSTM layer:

    ```python
    x = Masking(mask_value=0., input_shape=(timesteps, features))(x)
    x = LSTM(32)(x)
    ```

    Polyaxonfile usage:

    ```yaml
    Masking:
      mask_value: 0
    ```
    """
    IDENTIFIER = 'Masking'
    SCHEMA = MaskingSchema

    def __init__(self, mask_value=0., **kwargs):
        super(MaskingConfig, self).__init__(**kwargs)
        self.mask_value = mask_value


class DropoutSchema(BaseLayerSchema):
    rate = fields.Float(validate=validate.Range(0, 1))
    noise_shape = fields.List(fields.Int(), default=None, missing=None)
    seed = fields.Int(default=None, missing=None)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return DropoutConfig(**data)

    @post_dump
    def unmake(self, data):
        return DropoutConfig.remove_reduced_attrs(data)


class DropoutConfig(BaseLayerConfig):
    """Applies Dropout to the input.

    Dropout consists in randomly setting
    a fraction `rate` of input units to 0 at each update during training time,
    which helps prevent overfitting.

    Args:
        rate: float between 0 and 1. Fraction of the input units to drop.
        noise_shape: 1D integer tensor representing the shape of the
            binary dropout mask that will be multiplied with the input.
            For instance, if your inputs have shape
            `(batch_size, timesteps, features)` and
            you want the dropout mask to be the same for all timesteps,
            you can use `noise_shape=(batch_size, 1, features)`.
        seed: A Python integer to use as random seed.

    Polyaxonfile usage:

    ```yaml
    Dropout:
      rate: 0.5
    ```
    """
    IDENTIFIER = 'Dropout'
    SCHEMA = DropoutSchema

    def __init__(self, rate, noise_shape=None, seed=None, **kwargs):
        super(DropoutConfig, self).__init__(**kwargs)
        self.rate = rate
        self.noise_shape = noise_shape
        self.seed = seed


class SpatialDropout1DSchema(DropoutSchema):
    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return SpatialDropout1DConfig(**data)

    @post_dump
    def unmake(self, data):
        return SpatialDropout1DConfig.remove_reduced_attrs(data)


class SpatialDropout1DConfig(DropoutConfig):
    """Spatial 1D version of Dropout.

    This version performs the same function as Dropout, however it drops
    entire 1D feature maps instead of individual elements. If adjacent frames
    within feature maps are strongly correlated (as is normally the case in
    early convolution layers) then regular dropout will not regularize the
    activations and will otherwise just result in an effective learning rate
    decrease. In this case, SpatialDropout1D will help promote independence
    between feature maps and should be used instead.

    Args:
        rate: float between 0 and 1. Fraction of the input units to drop.

    Input shape:
        3D tensor with shape:
        `(samples, timesteps, channels)`

    Output shape:
        Same as input

    References:
        - [Efficient Object Localization Using Convolutional
          Networks](https://arxiv.org/abs/1411.4280)

    Polyaxonfile usage:

    ```yaml
    SpatialDropout1D:
      rate: 0.5
    ```
    """
    IDENTIFIER = 'SpatialDropout1D'
    SCHEMA = SpatialDropout1DSchema


class SpatialDropout2DSchema(DropoutSchema):
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return SpatialDropout2DConfig(**data)

    @post_dump
    def unmake(self, data):
        return SpatialDropout2DConfig.remove_reduced_attrs(data)


class SpatialDropout2DConfig(DropoutConfig):
    """Spatial 2D version of Dropout.

    This version performs the same function as Dropout, however it drops
    entire 2D feature maps instead of individual elements. If adjacent pixels
    within feature maps are strongly correlated (as is normally the case in
    early convolution layers) then regular dropout will not regularize the
    activations and will otherwise just result in an effective learning rate
    decrease. In this case, SpatialDropout2D will help promote independence
    between feature maps and should be used instead.

    Args:
        rate: float between 0 and 1. Fraction of the input units to drop.
        data_format: 'channels_first' or 'channels_last'.
            In 'channels_first' mode, the channels dimension
            (the depth) is at index 1,
            in 'channels_last' mode is it at index 3.
            If you never set it, then it will be "channels_last".

    Input shape:
        4D tensor with shape:
        `(samples, channels, rows, cols)` if data_format='channels_first'
        or 4D tensor with shape:
        `(samples, rows, cols, channels)` if data_format='channels_last'.

    Output shape:
        Same as input

    References:
        - [Efficient Object Localization Using Convolutional
          Networks](https://arxiv.org/abs/1411.4280)

    Polyaxonfile usage:

    ```yaml
    SpatialDropout2D:
      rate: 0.5
    ```
    """
    IDENTIFIER = 'SpatialDropout2D'
    SCHEMA = SpatialDropout2DSchema

    def __init__(self, rate, data_format=None, **kwargs):
        super(SpatialDropout2DConfig, self).__init__(rate, **kwargs)
        self.data_format = data_format


class SpatialDropout3DSchema(DropoutSchema):
    data_format = fields.Str(default=None, missing=None,
                             validate=validate.OneOf('channels_first', 'channels_last'))

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return SpatialDropout3DConfig(**data)

    @post_dump
    def unmake(self, data):
        return SpatialDropout3DConfig.remove_reduced_attrs(data)


class SpatialDropout3DConfig(DropoutConfig):
    """Spatial 3D version of Dropout.

    This version performs the same function as Dropout, however it drops
    entire 3D feature maps instead of individual elements. If adjacent voxels
    within feature maps are strongly correlated (as is normally the case in
    early convolution layers) then regular dropout will not regularize the
    activations and will otherwise just result in an effective learning rate
    decrease. In this case, SpatialDropout3D will help promote independence
    between feature maps and should be used instead.

    Args:
        rate: float between 0 and 1. Fraction of the input units to drop.
        data_format: 'channels_first' or 'channels_last'.
            In 'channels_first' mode, the channels dimension (the depth)
            is at index 1, in 'channels_last' mode is it at index 4.
            If you never set it, then it will be "channels_last".

    Input shape:
        5D tensor with shape:
        `(samples, channels, dim1, dim2, dim3)` if data_format='channels_first'
        or 5D tensor with shape:
        `(samples, dim1, dim2, dim3, channels)` if data_format='channels_last'.

    Output shape:
        Same as input

    References:
        - [Efficient Object Localization Using Convolutional
          Networks](https://arxiv.org/abs/1411.4280)

    Polyaxonfile usage:

    ```yaml
    SpatialDropout3D:
      rate: 0.5
    ```
    """
    IDENTIFIER = 'SpatialDropout3D'
    SCHEMA = SpatialDropout3DSchema

    def __init__(self, rate, data_format=None, **kwargs):
        super(SpatialDropout3DConfig, self).__init__(rate, **kwargs)
        self.data_format = data_format


class ActivationSchema(BaseLayerSchema):
    activation = StrOrFct(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return ActivationConfig(**data)

    @post_dump
    def unmake(self, data):
        return ActivationConfig.remove_reduced_attrs(data)


class ActivationConfig(BaseLayerConfig):
    """Applies an activation function to an output.

    Args:
        activation: name of activation function.

    Input shape:
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    Output shape:
        Same shape as input.

    Polyaxonfile usage:

    ```yaml
    Activation:
      activation: tanh
    ```
    """
    IDENTIFIER = 'Activation'
    SCHEMA = ActivationSchema

    def __init__(self, activation, **kwargs):
        super(ActivationConfig, self).__init__(**kwargs)
        self.activation = activation


class ReshapeSchema(BaseLayerSchema):
    target_shape = fields.List(fields.Int())

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return ReshapeConfig(**data)

    @post_dump
    def unmake(self, data):
        return ReshapeConfig.remove_reduced_attrs(data)


class ReshapeConfig(BaseLayerConfig):
    """Reshapes an output to a certain shape.

    Args:
        target_shape: target shape. Tuple of integers,
            does not include the samples dimension (batch size).

    Input shape:
        Arbitrary, although all dimensions in the input shaped must be fixed.
        Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    Output shape:
        `(batch_size,) + target_shape`

    Example:

    ```python
    # as first layer in a Sequential model
    x = Reshape((3, 4))(x)
    # now: x.output_shape == (None, 3, 4)
    # note: `None` is the batch dimension

    # also supports shape inference using `-1` as dimension
    x = Reshape((-1, 2, 2))(x)
    # now: x.output_shape == (None, 3, 2, 2)
    ```

    Polyaxonfile usage:

    ```yaml
    Reshape:
      target_shape: [-1, 2, 2]
    ```
    """
    IDENTIFIER = 'Reshape'
    SCHEMA = ReshapeSchema

    def __init__(self, target_shape, **kwargs):
        super(ReshapeConfig, self).__init__(**kwargs)
        self.target_shape = target_shape


class PermuteSchema(BaseLayerSchema):
    dims = fields.List(fields.Int())

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return PermuteConfig(**data)

    @post_dump
    def unmake(self, data):
        return PermuteConfig.remove_reduced_attrs(data)


class PermuteConfig(BaseLayerConfig):
    """Permutes the dimensions of the input according to a given pattern.

    Useful for e.g. connecting RNNs and convnets together.

    Args:
        dims: Tuple of integers. Permutation pattern, does not include the
            samples dimension. Indexing starts at 1.
            For instance, `(2, 1)` permutes the first and second dimension
            of the input.

    Input shape:
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    Output shape:
        Same as the input shape, but with the dimensions re-ordered according
        to the specified pattern.

    Example:

    ```python
    x = Permute((2, 1), input_shape=(10, 64))(x)
    # now: X.output_shape == (None, 64, 10)
    # note: `None` is the batch dimension
    ```

    Polyaxonfile usage:

    ```yaml
    Reshape:
      target_shape: [-1, 2, 2]
    ```
    """
    IDENTIFIER = 'Permute'
    SCHEMA = PermuteSchema

    def __init__(self, dims, **kwargs):
        super(PermuteConfig, self).__init__(**kwargs)
        self.dims = dims


class FlattenSchema(BaseLayerSchema):
    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return FlattenConfig(**data)

    @post_dump
    def unmake(self, data):
        return FlattenConfig.remove_reduced_attrs(data)


class FlattenConfig(BaseLayerConfig):
    """Flattens the input. Does not affect the batch size.

    Example:

    ```python
    x = Convolution2D(64, 3, 3,
                     border_mode='same',
                     input_shape=(3, 32, 32))(x)
    # now: x.output_shape == (None, 64, 32, 32)

    x = Flatten()(x)
    # now: x.output_shape == (None, 65536)
    ```

    Polyaxonfile usage:

    ```yaml
    Flatten:
    ```
    """
    IDENTIFIER = 'Flatten'
    SCHEMA = FlattenSchema


class RepeatVectorSchema(BaseLayerSchema):
    n = fields.Int()

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return RepeatVectorConfig(**data)

    @post_dump
    def unmake(self, data):
        return RepeatVectorConfig.remove_reduced_attrs(data)


class RepeatVectorConfig(BaseLayerConfig):
    """Repeats the input n times.

    Example:

    ```python
    x = Dense(32)(x)
    # now: x.output_shape == (None, 32)
    # note: `None` is the batch dimension

    x = RepeatVector(3)(x)
    # now: x.output_shape == (None, 3, 32)
    ```

    Args:
        n: integer, repetition factor.

    Input shape:
        2D tensor of shape `(num_samples, features)`.

    Output shape:
        3D tensor of shape `(num_samples, n, features)`.

    Polyaxonfile usage:

    ```yaml
    RepeatVector:
      n: 32
    ```
    """
    IDENTIFIER = 'RepeatVector'
    SCHEMA = RepeatVectorSchema

    def __init__(self, n, **kwargs):
        super(RepeatVectorConfig, self).__init__(**kwargs)
        self.n = n


# class LambdaSchema(BaseLayerSchema):


class DenseSchema(BaseLayerSchema):
    units = fields.Int()
    activation = StrOrFct(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    use_bias = fields.Bool(allow_none=True)
    kernel_initializer = fields.Nested(InitializerSchema, allow_none=True)
    bias_initializer = fields.Nested(InitializerSchema, allow_none=True)
    kernel_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    bias_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    activity_regularizer = fields.Nested(RegularizerSchema, allow_none=True)
    kernel_constraint = fields.Nested(ConstraintSchema, allow_none=True)
    bias_constraint = fields.Nested(ConstraintSchema, allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return DenseConfig(**data)

    @post_dump
    def unmake(self, data):
        return DenseConfig.remove_reduced_attrs(data)


class DenseConfig(BaseLayerConfig):
    """Just your regular densely-connected NN layer.

    `Dense` implements the operation:
    `output = activation(dot(input, kernel) + bias)`
    where `activation` is the element-wise activation function
    passed as the `activation` argument, `kernel` is a weights matrix
    created by the layer, and `bias` is a bias vector created by the layer
    (only applicable if `use_bias` is `True`).

    Note: if the input to the layer has a rank greater than 2, then
    it is flattened prior to the initial dot product with `kernel`.

    Example:

    ```python
    # as first layer in a sequential model:
    x = Dense(32)(x)
    # now the model will take as input arrays of shape (*, 16)
    # and output arrays of shape (*, 32)
    ```

    Args:
        units: Positive integer, dimensionality of the output space.
        activation: Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix.
        bias_initializer: Initializer for the bias vector.
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
        kernel_constraint: Constraint function applied to
            the `kernel` weights matrix.
        bias_constraint: Constraint function applied to the bias vector.

    Input shape:
        nD tensor with shape: `(batch_size, ..., input_dim)`.
        The most common situation would be
        a 2D input with shape `(batch_size, input_dim)`.

    Output shape:
        nD tensor with shape: `(batch_size, ..., units)`.
        For instance, for a 2D input with shape `(batch_size, input_dim)`,
        the output would have shape `(batch_size, units)`.

    Polyaxonfile usage:

    ```yaml
    Dense:
      units: 32
      activation: sigmoid
    ```
    """
    IDENTIFIER = 'Dense'
    SCHEMA = DenseSchema

    def __init__(self,
                 units,
                 activation=None,
                 use_bias=True,
                 kernel_initializer=GlorotNormalInitializerConfig(),
                 bias_initializer=ZerosInitializerConfig(),
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        super(DenseConfig, self).__init__(**kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer
        self.activity_regularizer = activity_regularizer
        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint


class ActivityRegularizationSchema(BaseLayerSchema):
    l1 = fields.Float(allow_none=True)
    l2 = fields.Float(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return ActivityRegularizationConfig(**data)

    @post_dump
    def unmake(self, data):
        return ActivityRegularizationConfig.remove_reduced_attrs(data)


class ActivityRegularizationConfig(BaseLayerConfig):
    """Layer that applies an update to the cost function based input activity.

    Args:
        l1: L1 regularization factor (positive float).
        l2: L2 regularization factor (positive float).

    Input shape:
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    Output shape:
        Same shape as input.

    Polyaxonfile usage:

    ```yaml
    ActivityRegularization:
      l1: 0.1
      l2: 0.2
    ```
    """
    IDENTIFIER = 'ActivityRegularization'
    SCHEMA = ActivityRegularizationSchema

    def __init__(self, l1=0., l2=0., **kwargs):
        super(ActivityRegularizationConfig, self).__init__(**kwargs)
        self.l1 = l1
        self.l2 = l2


class CastSchema(BaseLayerSchema):
    dtype = DType()

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return CastConfig(**data)

    @post_dump
    def unmake(self, data):
        return CastConfig.remove_reduced_attrs(data)


class CastConfig(BaseLayerConfig):
    """Casts a tensor to a new type.

    The operation casts `x` (in case of `Tensor`) or `x.values`
    (in case of `SparseTensor`) to `dtype`.

    For example:

    ```python
    x = tf.constant([1.8, 2.2], dtype=tf.float32)
    x = Cast(dtype=tf.int32)(x)  # [1, 2], dtype=tf.int32
    ```

    Args:
      x: A `Tensor` or `SparseTensor`.
      dtype: The destination type.
      name: A name for the operation (optional).

    Returns:
      A `Tensor` or `SparseTensor` with same shape as `x`.

    Raises:
      TypeError: If `x` cannot be cast to the `dtype`.

    Polyaxonfile usage:

    ```yaml
    Cast:
      dtype: float32
    ```
    """
    IDENTIFIER = 'Cast'
    SCHEMA = CastSchema

    def __init__(self, dtype, **kwargs):
        super(CastConfig, self).__init__(**kwargs)
        self.dtype = dtype



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema


class UserSchema(BaseSchema):
    username = fields.Str()
    email = fields.Email(allow_none=True)
    is_superuser = fields.Bool(allow_none=True)

    @staticmethod
    def schema_config():
        return UserConfig


class UserConfig(BaseConfig):
    SCHEMA = UserSchema
    IDENTIFIER = 'user'

    def __init__(self, username, email, is_superuser=False):
        self.username = username
        self.email = email
        self.is_superuser = is_superuser



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import ValidationError, fields, validates_schema

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.utils import UUID, FloatOrStr, IntOrStr


class TensorflowClusterSchema(BaseSchema):
    master = fields.List(fields.Str(), allow_none=True)
    worker = fields.List(fields.Str(), allow_none=True)
    ps = fields.List(fields.Str(), allow_none=True)

    @staticmethod
    def schema_config():
        return TensorflowClusterConfig


class TensorflowClusterConfig(BaseConfig):
    """
    Tensorflow cluster config.

    Args:
        master: list(str). The list of master host in a tensorflow cluster.
        worker: list(str). The list of worker host in a tensorflow cluster.
        ps: list(str). The list of ps host in a tensorflow cluster.
    """
    IDENTIFIER = 'tensorflow_cluster'
    SCHEMA = TensorflowClusterSchema

    def __init__(self, master=None, worker=None, ps=None):
        self.master = master
        self.worker = worker
        self.ps = ps


class HorovodClusterSchema(BaseSchema):
    master = fields.List(fields.Str(), allow_none=True)
    worker = fields.List(fields.Str(), allow_none=True)

    @staticmethod
    def schema_config():
        return HorovodClusterConfig


class HorovodClusterConfig(BaseConfig):
    """
    Horovod cluster config.

    Args:
        master: list(str). The list of master host in a Horovod cluster.
        worker: list(str). The list of worker host in a Horovod cluster.
    """
    IDENTIFIER = 'horovod_cluster'
    SCHEMA = HorovodClusterSchema

    def __init__(self, master=None, worker=None):
        self.master = master
        self.worker = worker


class PytorchClusterSchema(BaseSchema):
    master = fields.List(fields.Str(), allow_none=True)
    worker = fields.List(fields.Str(), allow_none=True)

    @staticmethod
    def schema_config():
        return PytorchClusterConfig


class PytorchClusterConfig(BaseConfig):
    """
    Pytorch cluster config.

    Args:
        master: list(str). The list of master host in a Pytorch cluster.
        worker: list(str). The list of worker host in a Pytorch cluster.
    """
    IDENTIFIER = 'pytorch_cluster'
    SCHEMA = PytorchClusterSchema

    def __init__(self, master=None, worker=None):
        self.master = master
        self.worker = worker


class MXNetClusterSchema(BaseSchema):
    master = fields.List(fields.Str(), allow_none=True)
    worker = fields.List(fields.Str(), allow_none=True)
    server = fields.List(fields.Str(), allow_none=True)

    @staticmethod
    def schema_config():
        return MXNetClusterConfig


class MXNetClusterConfig(BaseConfig):
    """
    MXNet cluster config.

    Args:
        master: list(str). The list of master host in a tensorflow cluster.
        worker: list(str). The list of worker host in a tensorflow cluster.
        server: list(str). The list of server host in a tensorflow cluster.
    """
    IDENTIFIER = 'mxnet_cluster'
    SCHEMA = MXNetClusterSchema

    def __init__(self, master=None, worker=None, server=None):
        self.master = master
        self.worker = worker
        self.server = server


class K8SResourcesEntrySchema(BaseSchema):
    cpu = FloatOrStr(allow_none=True)
    memory = IntOrStr(allow_none=True)
    gpu = fields.Int(allow_none=True)
    tpu = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return K8SResourcesEntryConfig


class K8SResourcesEntryConfig(BaseConfig):
    SCHEMA = K8SResourcesEntrySchema
    REDUCED_ATTRIBUTES = ['cpu', 'memory', 'gpu', 'tpu']

    def __init__(self, cpu=None, memory=None, gpu=None, tpu=None):
        self.cpu = cpu
        self.memory = memory
        self.gpu = gpu
        self.tpu = tpu


class K8SContainerResourcesSchema(BaseSchema):
    limits = fields.Nested(K8SResourcesEntrySchema, allow_none=True)
    requests = fields.Nested(K8SResourcesEntrySchema, allow_none=True)

    @staticmethod
    def schema_config():
        return K8SContainerResourcesConfig


class K8SContainerResourcesConfig(BaseConfig):
    """
    K8S container resources config.

    Args:
        limits: `K8SResourcesEntry`.
        requests: `K8SResourcesEntry`.
    """
    IDENTIFIER = 'resources'
    SCHEMA = K8SContainerResourcesSchema
    REDUCED_ATTRIBUTES = ['limits', 'requests']

    def __init__(self, limits=None, requests=None):
        self.limits = limits
        self.requests = requests


class K8SResourcesSchema(BaseSchema):
    limits = fields.Float(allow_none=True)
    requests = fields.Float(allow_none=True)

    @staticmethod
    def schema_config():
        return K8SResourcesConfig


class K8SResourcesConfig(BaseConfig):
    """
    K8S resources config.

    Args:
        limits: `float`.
        requests: `float`.
    """
    IDENTIFIER = 'resources'
    SCHEMA = K8SResourcesSchema

    def __init__(self, limits=None, requests=None):
        self.limits = limits
        self.requests = requests

    def __add__(self, other):
        if not other:
            return self

        if self.requests:
            if other.requests:
                self.requests += other.requests
        elif other.requests:
            self.requests = other.requests

        if self.limits:
            if other.limits:
                self.limits += other.limits
        elif other.limits:
            self.limits = other.limits

        return self


class PodResourcesSchema(BaseSchema):
    cpu = fields.Nested(K8SResourcesSchema, allow_none=True)
    memory = fields.Nested(K8SResourcesSchema, allow_none=True)
    gpu = fields.Nested(K8SResourcesSchema, allow_none=True)
    tpu = fields.Nested(K8SResourcesSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return PodResourcesConfig


class PodResourcesConfig(BaseConfig):
    """
    Pod resources config.

    Args:
        cpu: `K8SResourcesConfig`.
        memory: `K8SResourcesConfig`.
        gpu: `K8SResourcesConfig`.
        tpu: `K8SResourcesConfig`.
    """
    IDENTIFIER = 'pod_resources'
    SCHEMA = PodResourcesSchema

    def __init__(self, cpu=None, memory=None, gpu=None, tpu=None):
        self.cpu = cpu
        self.memory = memory
        self.gpu = gpu
        self.tpu = tpu

    def __add__(self, other):
        if not other:
            return self

        if self.cpu:
            if other.cpu:
                self.cpu += other.cpu
        elif other.cpu:
            self.cpu = K8SResourcesConfig()
            self.cpu += other.cpu

        if self.memory:
            if other.memory:
                self.memory += other.memory
        elif other.memory:
            self.memory = K8SResourcesConfig()
            self.memory += other.memory

        if self.gpu:
            if other.gpu:
                self.gpu += other.gpu
        elif other.gpu:
            self.gpu = K8SResourcesConfig()
            self.gpu += other.gpu

        if self.tpu:
            if other.tpu:
                self.tpu += other.tpu
        elif other.gpu:
            self.tpu = K8SResourcesConfig()
            self.tpu += other.tpu
        return self


class GPUOptionsSchema(BaseSchema):
    gpu_memory_fraction = fields.Float(allow_none=True)
    allow_growth = fields.Bool(allow_none=True)
    per_process_gpu_memory_fraction = fields.Float(allow_none=True)

    @staticmethod
    def schema_config():
        return GPUOptionsConfig


class GPUOptionsConfig(BaseConfig):
    IDENTIFIER = 'session'
    SCHEMA = GPUOptionsSchema

    def __init__(self,
                 gpu_memory_fraction=None,
                 allow_growth=True,
                 per_process_gpu_memory_fraction=None):
        self.gpu_memory_fraction = gpu_memory_fraction
        self.allow_growth = allow_growth
        self.per_process_gpu_memory_fraction = per_process_gpu_memory_fraction


class SessionSchema(BaseSchema):
    gpu_options = fields.Nested(GPUOptionsSchema, allow_none=True)
    log_device_placement = fields.Bool(allow_none=True)
    allow_soft_placement = fields.Bool(allow_none=True)
    intra_op_parallelism_threads = fields.Int(allow_none=True)
    inter_op_parallelism_threads = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return SessionConfig


class SessionConfig(BaseConfig):
    IDENTIFIER = 'session'
    SCHEMA = SessionSchema

    def __init__(self,
                 log_device_placement=True,
                 gpu_options=GPUOptionsConfig(),
                 allow_soft_placement=True,
                 intra_op_parallelism_threads=None,
                 inter_op_parallelism_threads=None):
        self.gpu_options = gpu_options
        self.log_device_placement = log_device_placement
        self.allow_soft_placement = allow_soft_placement
        self.intra_op_parallelism_threads = intra_op_parallelism_threads
        self.inter_op_parallelism_threads = inter_op_parallelism_threads


class TFRunSchema(BaseSchema):
    tf_random_seed = fields.Int(allow_none=True)
    save_summary_steps = fields.Int(allow_none=True)
    save_checkpoints_secs = fields.Int(allow_none=True)
    save_checkpoints_steps = fields.Int(allow_none=True)
    keep_checkpoint_max = fields.Int(allow_none=True)
    keep_checkpoint_every_n_hours = fields.Int(allow_none=True)

    session = fields.Nested(SessionSchema, allow_none=True)
    cluster = fields.Nested(TensorflowClusterSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return TFRunConfig


class TFRunConfig(BaseConfig):
    IDENTIFIER = 'run'
    SCHEMA = TFRunSchema

    def __init__(self,
                 tf_random_seed=None,
                 save_summary_steps=100,
                 save_checkpoints_secs=None,
                 save_checkpoints_steps=None,
                 keep_checkpoint_max=5,
                 keep_checkpoint_every_n_hours=10000,
                 session=None,
                 cluster=None):
        self.tf_random_seed = tf_random_seed
        self.save_summary_steps = save_summary_steps
        self.save_checkpoints_secs = save_checkpoints_secs
        self.save_checkpoints_steps = save_checkpoints_steps
        self.keep_checkpoint_max = keep_checkpoint_max
        self.keep_checkpoint_every_n_hours = keep_checkpoint_every_n_hours
        self.session = session
        self.cluster = cluster


class PodEnvironmentSchema(BaseSchema):
    # To indicate which worker/ps index this session config belongs to
    index = fields.Int(allow_none=True)
    resources = fields.Nested(PodResourcesSchema, allow_none=True)
    node_selector = fields.Dict(allow_none=True)
    affinity = fields.Dict(allow_none=True)
    tolerations = fields.List(fields.Dict(), allow_none=True)

    @staticmethod
    def schema_config():
        return PodEnvironmentConfig


class PodEnvironmentConfig(BaseConfig):
    """
    Pod environment config.

    Args:
        index: `int | None`. The index of the pod.
        resources: `PodResourcesConfig`.
        node_selector: `dict`.
        affinity: `dict`.
        tolerations: `list(dict)`.
    """
    IDENTIFIER = 'pod_environment'
    SCHEMA = PodEnvironmentSchema
    REDUCED_ATTRIBUTES = ['index', 'resources', 'node_selector', 'affinity', 'tolerations']

    def __init__(self,
                 index=None,
                 resources=None,
                 node_selector=None,
                 affinity=None,
                 tolerations=None,
                 ):
        self.index = index
        self.resources = resources
        self.node_selector = node_selector
        self.affinity = affinity
        self.tolerations = tolerations


class TensorflowPodEnvironmentSchema(PodEnvironmentSchema):
    config = fields.Nested(SessionSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return TensorflowPodEnvironmentConfig


class TensorflowPodEnvironmentConfig(PodEnvironmentConfig):
    IDENTIFIER = 'pod_environment'
    SCHEMA = TensorflowPodEnvironmentSchema
    REDUCED_ATTRIBUTES = PodEnvironmentConfig.REDUCED_ATTRIBUTES + ['config']

    def __init__(self,
                 index=None,
                 config=None,
                 resources=None,
                 node_selector=None,
                 affinity=None,
                 tolerations=None,
                 ):
        self.config = config
        super(TensorflowPodEnvironmentConfig, self).__init__(
            index=index,
            resources=resources,
            node_selector=node_selector,
            affinity=affinity,
            tolerations=tolerations,
        )


class FrameworkEnvironmentMixin(object):
    @staticmethod
    def _get_env_indexed_property(obj, getter):
        if not obj:
            return {}
        return {o.index: getter(o) for o in obj if getter(o)}

    @property
    def default_worker_config(self):
        return self.default_worker.config if self.default_worker else None

    @property
    def default_worker_resources(self):
        return self.default_worker.resources if self.default_worker else None

    @property
    def default_worker_node_selector(self):
        return self.default_worker.node_selector if self.default_worker else None

    @property
    def default_worker_affinity(self):
        return self.default_worker.affinity if self.default_worker else None

    @property
    def default_worker_tolerations(self):
        return self.default_worker.tolerations if self.default_worker else None

    @property
    def default_ps_config(self):
        return self.default_ps.config if self.default_ps else None

    @property
    def default_ps_resources(self):
        return self.default_ps.resources if self.default_ps else None

    @property
    def default_ps_node_selector(self):
        return self.default_ps.node_selector if self.default_ps else None

    @property
    def default_ps_affinity(self):
        return self.default_ps.affinity if self.default_ps else None

    @property
    def default_ps_tolerations(self):
        return self.default_ps.tolerations if self.default_ps else None

    @property
    def worker_configs(self):
        return self._get_env_indexed_property(obj=self.worker, getter=lambda o: o.config)

    @property
    def worker_resources(self):
        return self._get_env_indexed_property(obj=self.worker, getter=lambda o: o.resources)

    @property
    def worker_node_selectors(self):
        return self._get_env_indexed_property(obj=self.worker, getter=lambda o: o.node_selector)

    @property
    def worker_affinities(self):
        return self._get_env_indexed_property(obj=self.worker, getter=lambda o: o.affinity)

    @property
    def worker_tolerations(self):
        return self._get_env_indexed_property(obj=self.worker, getter=lambda o: o.tolerations)

    @property
    def ps_configs(self):
        return self._get_env_indexed_property(obj=self.ps, getter=lambda o: o.config)

    @property
    def ps_resources(self):
        return self._get_env_indexed_property(obj=self.ps, getter=lambda o: o.resources)

    @property
    def ps_node_selectors(self):
        return self._get_env_indexed_property(obj=self.ps, getter=lambda o: o.node_selector)

    @property
    def ps_affinities(self):
        return self._get_env_indexed_property(obj=self.ps, getter=lambda o: o.affinity)

    @property
    def ps_tolerations(self):
        return self._get_env_indexed_property(obj=self.ps, getter=lambda o: o.tolerations)


class TensorflowSchema(BaseSchema):
    n_workers = fields.Int(allow_none=True)
    n_ps = fields.Int(allow_none=True)
    delay_workers_by_global_step = fields.Bool(allow_none=True)
    run_config = fields.Nested(TFRunSchema, allow_none=True)
    default_worker = fields.Nested(TensorflowPodEnvironmentSchema, allow_none=True)
    default_ps = fields.Nested(TensorflowPodEnvironmentSchema, allow_none=True)
    worker = fields.Nested(TensorflowPodEnvironmentSchema, many=True, allow_none=True)
    ps = fields.Nested(TensorflowPodEnvironmentSchema, many=True, allow_none=True)

    @staticmethod
    def schema_config():
        return TensorflowConfig


class TensorflowConfig(BaseConfig, FrameworkEnvironmentMixin):
    """
    Tensorflow environment config.

    Args:
        n_workers: `int`. The number of workers requested for training the model.
        n_ps: `int`. The number of ps requested for training the model.
        default_worker: `PodEnvironment`. The default pod environment to use for all workers.
        default_ps: `PodEnvironment`. The default pod environment to use for all ps.
        worker: `list(PodEnvironment)`. The pod environment with index specified to use
            for the specific worker.
        ps: `list(PodEnvironment)`. The pod environment with index specified to use
            for the specific ps.
    """
    IDENTIFIER = 'tensorflow'
    SCHEMA = TensorflowSchema

    def __init__(self,
                 n_workers=0,
                 n_ps=0,
                 delay_workers_by_global_step=False,
                 run_config=None,
                 default_worker=None,
                 default_ps=None,
                 worker=None,
                 ps=None,
                 ):
        self.n_workers = n_workers
        self.n_ps = n_ps
        self.delay_workers_by_global_step = delay_workers_by_global_step
        self.run_config = run_config
        self.default_worker = default_worker
        self.default_ps = default_ps
        self.worker = worker
        self.ps = ps


class HorovodSchema(BaseSchema):
    n_workers = fields.Int(allow_none=True)
    default_worker = fields.Nested(PodEnvironmentSchema, allow_none=True)
    worker = fields.Nested(PodEnvironmentSchema, many=True, allow_none=True)

    @staticmethod
    def schema_config():
        return HorovodConfig


class HorovodConfig(BaseConfig, FrameworkEnvironmentMixin):
    """
    Horovod environment config.

    Args:
        n_workers: `int`. The number of workers requested for training the model.
        default_worker: `PodEnvironment`. The default pod environment to use for all workers.
        worker: `list(PodEnvironment)`. The pod environment with index specified to use
            for the specific worker.
    """
    IDENTIFIER = 'horovod'
    SCHEMA = HorovodSchema

    def __init__(self,
                 n_workers=0,
                 default_worker=None,
                 worker=None,
                 ):
        self.n_workers = n_workers
        self.default_worker = default_worker
        self.worker = worker


class PytorchSchema(BaseSchema):
    n_workers = fields.Int(allow_none=True)
    default_worker = fields.Nested(PodEnvironmentSchema, allow_none=True)
    worker = fields.Nested(PodEnvironmentSchema, many=True, allow_none=True)

    @staticmethod
    def schema_config():
        return PytorchConfig


class PytorchConfig(BaseConfig, FrameworkEnvironmentMixin):
    """
    Pytorch environment config.

    Args:
        n_workers: `int`. The number of workers requested for training the model.
        default_worker: `PodEnvironment`. The default pod environment to use for all workers.
        worker: `list(PodEnvironment)`. The pod environment with index specified to use
            for the specific worker.
    """
    IDENTIFIER = 'pytorch'
    SCHEMA = PytorchSchema

    def __init__(self,
                 n_workers=0,
                 default_worker=None,
                 worker=None,
                 ):
        self.n_workers = n_workers
        self.default_worker = default_worker
        self.worker = worker


class MXNetSchema(BaseSchema):
    n_workers = fields.Int(allow_none=True)
    n_ps = fields.Int(allow_none=True)
    default_worker = fields.Nested(PodEnvironmentSchema, allow_none=True)
    default_ps = fields.Nested(PodEnvironmentSchema, allow_none=True)
    worker = fields.Nested(PodEnvironmentSchema, many=True, allow_none=True)
    ps = fields.Nested(PodEnvironmentSchema, many=True, allow_none=True)

    @staticmethod
    def schema_config():
        return MXNetConfig


class MXNetConfig(BaseConfig, FrameworkEnvironmentMixin):
    """
    MXNet environment config.

    Args:
        n_workers: `int`. The number of workers requested for training the model.
        n_ps: `int`. The number of ps requested for training the model.
        default_worker: `PodEnvironment`. The default pod environment to use for all workers.
        default_ps: `PodEnvironment`. The default pod environment to use for all ps.
        worker: `list(PodEnvironment)`. The pod environment with index specified to use
            for the specific worker.
        ps: `list(PodEnvironment)`. The pod environment with index specified to use
            for the specific ps.
    """
    IDENTIFIER = 'mxnet'
    SCHEMA = MXNetSchema

    def __init__(self,
                 n_workers=0,
                 n_ps=0,
                 default_worker=None,
                 default_ps=None,
                 worker=None,
                 ps=None,
                 ):
        self.n_workers = n_workers
        self.n_ps = n_ps
        self.default_worker = default_worker
        self.default_ps = default_ps
        self.worker = worker
        self.ps = ps


def validate_frameworks(frameworks):
    if sum([1 for f in frameworks if f is not None]) > 1:
        raise ValidationError('Only one framework can be used.')


class PersistenceSchema(BaseSchema):
    data = fields.List(fields.Str(), allow_none=True)
    outputs = fields.Str(allow_none=True)

    @staticmethod
    def schema_config():
        return PersistenceConfig


class PersistenceConfig(BaseConfig):
    """
    Persistence config.

    Defines the list of persistent volumes to mount for this specific run.

    Args:
        data: `list(str)`. The list of the names of data persistence to mount.
        outputs: `list(str)`. The list of the names of outputs persistence to mount.
    """
    IDENTIFIER = 'persistence'
    SCHEMA = PersistenceSchema

    def __init__(self, data=None, outputs=None):
        self.data = data
        self.outputs = outputs


class OutputsSchema(BaseSchema):
    jobs = fields.List(IntOrStr(), allow_none=True)
    experiments = fields.List(IntOrStr(), allow_none=True)

    @staticmethod
    def schema_config():
        return OutputsConfig


class OutputsConfig(BaseConfig):
    """
    Outputs config.

    Defines the list of previous jobs/experiments outputs paths
    to make available to other experiments/jobs.

    Args:
        jobs: `list(str)`. The list of the names of jobs to make available for the current run.
        experiments: `list(str)`. The list of the names of experiments
            to make available for the current run.
    """
    IDENTIFIER = 'outputs'
    SCHEMA = OutputsSchema

    def __init__(self, jobs=None, experiments=None):
        self.jobs = jobs
        self.experiments = experiments


class EnvironmentSchema(PodEnvironmentSchema):
    cluster_uuid = UUID(allow_none=True)
    persistence = fields.Nested(PersistenceSchema, allow_none=True)
    outputs = fields.Nested(OutputsSchema, allow_none=True)
    secret_refs = fields.List(fields.Str(), allow_none=True)
    configmap_refs = fields.List(fields.Str(), allow_none=True)
    tensorflow = fields.Nested(TensorflowSchema, allow_none=True)
    horovod = fields.Nested(HorovodSchema, allow_none=True)
    mxnet = fields.Nested(MXNetSchema, allow_none=True)
    pytorch = fields.Nested(PytorchSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return EnvironmentConfig

    @validates_schema
    def validate_quantity(self, data):
        validate_frameworks([data.get('tensorflow'),
                             data.get('mxnet'),
                             data.get('pytorch'),
                             data.get('horovod')])


class EnvironmentConfig(PodEnvironmentConfig):
    """
    Environment config.

    Args:
        cluster_uuid: `str`. The cluster uuid.
        persistence: `PersistenceConfig`. The persistence config definition.
        outputs: `OutputsConfig`. The outputs config definition.
        resources: `PodResourcesConfig`. The resources config definition.
        node_selector: `dict`.
        affinity: `dict`.
        tolerations: `list(dict)`.
        tensorflow: `TensorflowConfig`.
        horovod: `HorovodConfig`.
        pytorch: `PytorchConfig`.
        mxnet: `MXNetConfig`.
    """
    IDENTIFIER = 'environment'
    SCHEMA = EnvironmentSchema

    def __init__(self,
                 cluster_uuid=None,
                 persistence=None,
                 outputs=None,
                 resources=None,
                 secret_refs=None,
                 configmap_refs=None,
                 node_selector=None,
                 affinity=None,
                 tolerations=None,
                 tensorflow=None,
                 horovod=None,
                 pytorch=None,
                 mxnet=None):
        self.cluster_uuid = cluster_uuid
        self.persistence = persistence
        self.outputs = outputs
        self.secret_refs = secret_refs
        self.configmap_refs = configmap_refs
        super(EnvironmentConfig, self).__init__(
            resources=resources,
            node_selector=node_selector,
            affinity=affinity,
            tolerations=tolerations,
        )
        validate_frameworks([tensorflow, horovod, pytorch, mxnet])
        self.tensorflow = tensorflow
        self.horovod = horovod
        self.pytorch = pytorch
        self.mxnet = mxnet



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import Schema, fields, post_dump, post_load

from polyaxon_schemas.base import BaseConfig
from polyaxon_schemas.ml.layers import LayerSchema
from polyaxon_schemas.utils import ObjectOrListObject, Tensor


class GraphSchema(Schema):
    input_layers = ObjectOrListObject(Tensor)
    output_layers = ObjectOrListObject(Tensor)
    layers = fields.Nested(LayerSchema, many=True)
    name = fields.Str(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return GraphConfig(**data)

    @post_dump
    def unmake(self, data):
        return GraphConfig.remove_reduced_attrs(data)


class GraphConfig(BaseConfig):
    """A Graph is a directed acyclic graph of layers.

    It is the topological form for creating a computational graph.
    The graph contains a list of layers/tensor operations, i.e.

    ```yaml
    - layer_name:
        arg1: value
        arg2: value
    - layer_name:
        arg1: value
        arg2: value
    ```

    The graph must define its inputs, i.e. `input_layers`,
    usually the names of the tensors that these graph is expecting from a preprocessing step.

    ```yaml
    graph:
      input_layers: ['input1', 'input2']
      layers:
        - layer_name:
        ...
    ```

    A graph can also specify the outputs, i.e. `output_layers`,
    when not specified, the `output_layers` will contain by default the last layer.
    when specified, if the `output_layers` do not contain the last layer it will be added.

    ```yaml
    graph:
      input_layers: ['image']
      layers:
        - Dense:
          units: 2
        - Dense:
          units: 2
          name: last_layer
    ```

    This is equivalent to

    ```yaml
    graph:
      input_layers: [image]
      layers:
        - Dense:
          units: 2
        - Dense:
          units: 2
          name: last_layer
      output_layers: [last_layer]
    ```

    Defining multiple outputs could be achieved in different ways,
    either by listing all `output_layers` or by using the `is_output: True` on the specific layer.

    ```yaml
    graph:
      input_layers: [image]
      layers:
        - Dense:
          units: 2
          is_output: true
        - Dense:
          units: 2
          name: last_layer
      output_layers: [last_layer]
    ```

    Means that, both `last_layer` and `dense_1` are in the list of the `output_layers`.

    This is the same as

    ```yaml
    graph:
      input_layers: [image]
      layers:
        - Dense:
          units: 2
        - Dense:
          units: 2
          name: last_layer
      output_layers: [dense_1, last_layer]
    ```

    A graph can optionally have a name

    ```yaml
    graph:
      name: cool_graph
      input_layers: [image]
      layers:
        - ...
    ```

    Args:
        name: the name of the graph
        input_layers: `str` or `list` of `str`, the input layers for the graph.
        output_layers: `str` or `list` of `str`, the output layers for the graph.
        layers: `list`, list of layers.

    Polyaxonfile usage:

    ```yaml
    graph:
      input_layers: image
      layers:
        - Dense:
            units: 128
        - Dense:
            units: 256
            name: dense_out
    ```

    or

    ```yaml
    graph:
      input_layers: image
      layers:
        - Dense:
            units: 128
        - Dense:
            units: 256
            name: dense_out
      output_layers: dense_out
    ```
    """
    SCHEMA = GraphSchema
    IDENTIFIER = 'graph'
    REDUCED_ATTRIBUTES = ['name']

    def __init__(self, input_layers, output_layers, layers, name='graph'):
        self.input_layers = input_layers
        self.output_layers = output_layers
        self.layers = layers
        self.name = name



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from hestia.humanize import humanize_timedelta
from marshmallow import fields, validate

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.environments import PodResourcesSchema
from polyaxon_schemas.utils import UUID


class JobSchema(BaseSchema):
    id = fields.Int(allow_none=True)
    uuid = UUID(allow_none=True)
    unique_name = fields.Str(allow_none=True)
    pod_id = fields.Str(allow_none=True)
    name = fields.Str(validate=validate.Regexp(regex=r'^[-a-zA-Z0-9_]+\Z'), allow_none=True)
    user = fields.Str(validate=validate.Regexp(regex=r'^[-a-zA-Z0-9_]+\Z'), allow_none=True)
    project = fields.Str(allow_none=True)
    build_job = fields.Str(allow_none=True)
    description = fields.Str(allow_none=True)
    tags = fields.List(fields.Str(), allow_none=True)
    last_status = fields.Str(allow_none=True)
    created_at = fields.LocalDateTime(allow_none=True)
    updated_at = fields.LocalDateTime(allow_none=True)
    started_at = fields.LocalDateTime(allow_none=True)
    finished_at = fields.LocalDateTime(allow_none=True)
    total_run = fields.Str(allow_none=True)
    is_clone = fields.Bool(allow_none=True)
    config = fields.Dict(allow_none=True)
    ttl = fields.Int(allow_none=True)
    resources = fields.Nested(PodResourcesSchema, allow_none=True)
    definition = fields.Dict(allow_none=True)

    @staticmethod
    def schema_config():
        return JobConfig


class JobConfig(BaseConfig):
    SCHEMA = JobSchema
    IDENTIFIER = 'Job'
    DEFAULT_INCLUDE_ATTRIBUTES = [
        'id', 'unique_name', 'user', 'last_status',
        'created_at', 'started_at', 'finished_at', 'total_run',
    ]
    DATETIME_ATTRIBUTES = ['created_at', 'updated_at', 'started_at', 'finished_at']

    def __init__(self,
                 id=None,  # pylint:disable=redefined-builtin
                 user=None,
                 uuid=None,
                 name=None,
                 unique_name=None,
                 pod_id=None,
                 project=None,
                 build_job=None,
                 description=None,
                 tags=None,
                 last_status=None,
                 definition=None,
                 created_at=None,
                 updated_at=None,
                 started_at=None,
                 finished_at=None,
                 is_clone=None,
                 config=None,
                 num_jobs=0,
                 resources=None,
                 ttl=None,
                 jobs=None,
                 total_run=None):
        self.id = id
        self.user = user
        self.uuid = uuid
        self.name = name
        self.unique_name = unique_name
        self.pod_id = pod_id
        self.project = project
        self.build_job = build_job
        self.description = description
        self.tags = tags
        self.last_status = last_status
        self.definition = definition
        self.started_at = self.localize_date(started_at)
        self.finished_at = self.localize_date(finished_at)
        self.created_at = self.localize_date(created_at)
        self.updated_at = self.localize_date(updated_at)
        self.is_clone = is_clone
        self.config = config  # The json compiled content of this experiment
        self.num_jobs = num_jobs
        self.resources = resources
        self.ttl = ttl
        self.jobs = jobs
        self.total_run = None
        if all([self.started_at, self.finished_at]):
            self.total_run = humanize_timedelta((self.finished_at - self.started_at).seconds)


class TensorboardJobSchema(JobSchema):
    experiment = fields.Int(allow_none=True)
    experiment_group = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return TensorboardJobConfig


class TensorboardJobConfig(JobConfig):
    SCHEMA = TensorboardJobSchema
    IDENTIFIER = 'TensorboardJob'
    DEFAULT_INCLUDE_ATTRIBUTES = [
        'id', 'unique_name', 'user', 'last_status', 'experiment', 'experiment_group',
        'created_at', 'started_at', 'finished_at', 'total_run', 'pod_id'
    ]

    def __init__(self,
                 id=None,  # pylint:disable=redefined-builtin
                 user=None,
                 uuid=None,
                 name=None,
                 unique_name=None,
                 pod_id=None,
                 project=None,
                 experiment=None,
                 experiment_group=None,
                 build_job=None,
                 description=None,
                 tags=None,
                 last_status=None,
                 definition=None,
                 created_at=None,
                 updated_at=None,
                 started_at=None,
                 finished_at=None,
                 is_clone=None,
                 config=None,
                 num_jobs=0,
                 resources=None,
                 ttl=None,
                 jobs=None,
                 total_run=None):
        super(TensorboardJobConfig, self).__init__(
            id=id,
            user=user,
            uuid=uuid,
            name=name,
            unique_name=unique_name,
            pod_id=pod_id,
            project=project,
            build_job=build_job,
            description=description,
            tags=tags,
            last_status=last_status,
            definition=definition,
            created_at=created_at,
            updated_at=updated_at,
            started_at=started_at,
            finished_at=finished_at,
            is_clone=is_clone,
            config=config,
            num_jobs=num_jobs,
            resources=resources,
            jobs=jobs,
            total_run=total_run,
        )
        self.experiment = experiment
        self.experiment_group = experiment_group


class JobStatusSchema(BaseSchema):
    id = fields.Int()
    uuid = UUID()
    job = fields.Int()
    created_at = fields.LocalDateTime()
    status = fields.Str()
    message = fields.Str(allow_none=True)
    details = fields.Dict(allow_none=True)

    @staticmethod
    def schema_config():
        return JobStatusConfig


class JobStatusConfig(BaseConfig):
    SCHEMA = JobStatusSchema
    IDENTIFIER = 'JobStatus'
    DATETIME_ATTRIBUTES = ['created_at']
    DEFAULT_EXCLUDE_ATTRIBUTES = ['job', 'uuid', 'details']

    def __init__(self,
                 id,  # pylint:disable=redefined-builtin
                 uuid,
                 job,
                 created_at,
                 status,
                 message=None,
                 details=None):
        self.id = id
        self.uuid = uuid
        self.job = job
        self.created_at = self.localize_date(created_at)
        self.status = status
        self.message = message
        self.details = details


class JobLabelSchema(BaseSchema):
    app = fields.Str(allow_none=True)
    project_name = fields.Str()
    experiment_group_name = fields.Str(allow_none=True)
    experiment_name = fields.Str(allow_none=True)
    job_name = fields.Str(allow_none=True)
    project_uuid = UUID()
    experiment_group_uuid = UUID(allow_none=True)
    experiment_uuid = UUID(allow_none=True)
    job_uuid = UUID(allow_none=True)
    task_type = fields.Str(allow_none=True)
    task_idx = fields.Str(allow_none=True)
    role = fields.Str()
    type = fields.Str()

    @staticmethod
    def schema_config():
        return JobLabelConfig


class JobLabelConfig(BaseConfig):
    SCHEMA = JobLabelSchema
    IDENTIFIER = 'JobLabel'
    REDUCED_ATTRIBUTES = [
        'app', 'experiment_name', 'job_name', 'experiment_uuid',
        'experiment_group_name', 'experiment_group_uuid',
        'task_type', 'task_idx', 'job_uuid'
    ]

    def __init__(self,
                 project_name,
                 project_uuid,
                 role,
                 type,  # pylint:disable=redefined-builtin
                 experiment_name=None,
                 experiment_uuid=None,
                 app=None,
                 task_type=None,
                 task_idx=None,
                 job_name=None,
                 job_uuid=None,
                 experiment_group_name=None,
                 experiment_group_uuid=None,
                 **kwargs):
        self.project_name = project_name
        self.experiment_group_name = experiment_group_name
        self.experiment_name = experiment_name
        self.project_uuid = project_uuid
        self.experiment_group_uuid = experiment_group_uuid
        self.experiment_uuid = experiment_uuid
        self.task_type = task_type
        self.task_idx = task_idx
        self.job_uuid = job_uuid
        self.job_name = job_name
        self.role = role
        self.type = type
        self.app = app



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_dict, assert_equal_graphs

from polyaxon_schemas.ml.bridges import NoOpBridgeConfig
from polyaxon_schemas.ml.losses import MeanSquaredErrorConfig, SoftmaxCrossEntropyConfig
from polyaxon_schemas.ml.metrics import AccuracyConfig, PrecisionConfig
from polyaxon_schemas.ml.models import (
    BaseModelConfig,
    ClassifierConfig,
    GeneratorConfig,
    RegressorConfig
)
from polyaxon_schemas.ml.optimizers import AdamConfig


class TestModelConfigs(TestCase):
    @staticmethod
    def assert_equal_models(result_model, expected_model):
        assert_equal_graphs(expected_model.pop('graph'), result_model.pop('graph'))
        assert_equal_dict(result_model, expected_model)

    def test_base_model_config(self):
        config_dict = {
            'graph': {
                'input_layers': ['image'],
                'output_layers': ['dense_0'],
                'layers': [
                    {
                        'Conv2D': {
                            'filters': 64,
                            'strides': [1, 1],
                            'kernel_size': [2, 2],
                            'activation': 'relu',
                            'name': 'convolution_1',
                        }
                    },
                    {'Dense': {'units': 17, 'name': 'dense_0'}}
                ]
            },
            'loss': SoftmaxCrossEntropyConfig(input_layer=['image', 0, 0],
                                              output_layer=['dense_0', 0, 0]).to_schema(),
            'optimizer': AdamConfig(learning_rate=0.01).to_schema(),
            'metrics': [
                AccuracyConfig(input_layer=['image', 0, 0],
                               output_layer=['dense_0', 0, 0]).to_schema(),
                PrecisionConfig(input_layer=['image', 0, 0],
                                output_layer=['dense_0', 0, 0]).to_schema(),
            ],
            'summaries': ['loss', 'gradients'],
            'clip_gradients': 0.5,
            'clip_embed_gradients': 0.,
            'name': 'model'}
        config = BaseModelConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        self.assert_equal_models(config_to_dict, config_dict)

    def test_classifier_model_config(self):
        config_dict = {
            'graph': {
                'input_layers': ['image'],
                'output_layers': ['dense_0'],
                'layers': [
                    {
                        'Conv2D': {
                            'filters': 64,
                            'strides': [1, 1],
                            'kernel_size': [2, 2],
                            'activation': 'relu',
                            'name': 'convolution_1',
                        }
                    },
                    {'Dense': {'units': 17, 'name': 'dense_0'}}
                ]
            },
            'one_hot_encode': True,
            'n_classes': 10,
            'loss': SoftmaxCrossEntropyConfig(['image', 0, 0],
                                              output_layer=['dense_0', 0, 0]).to_schema(),
            'optimizer': AdamConfig(learning_rate=0.01).to_schema(),
            'metrics': [
                AccuracyConfig(input_layer=['image', 0, 0],
                               output_layer=['dense_0', 0, 0]).to_schema(),
                PrecisionConfig(input_layer=['image', 0, 0],
                                output_layer=['dense_0', 0, 0]).to_schema(),
            ],
            'summaries': ['loss', 'gradients'],
            'clip_gradients': 0.5,
            'clip_embed_gradients': 0.,
            'name': 'model'}
        config = ClassifierConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        self.assert_equal_models(config_to_dict, config_dict)

    def test_regressor_model_config(self):
        config_dict = {
            'graph': {
                'input_layers': ['image'],
                'output_layers': ['dense_0'],
                'layers': [
                    {'LSTM': {'units': 5}},
                    {'Dense': {'units': 1, 'name': 'dense_0'}}
                ]
            },
            'loss': MeanSquaredErrorConfig(input_layer=['image', 0, 0],
                                           output_layer=['dense_0', 0, 0]).to_schema(),
            'optimizer': AdamConfig(learning_rate=0.01).to_schema(),
            'metrics': [],
            'summaries': ['loss', 'gradients'],
            'clip_gradients': 0.5,
            'clip_embed_gradients': 0.,
            'name': 'model'}
        config = RegressorConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        self.assert_equal_models(config_to_dict, config_dict)

    def test_generator_model_config(self):
        config_dict = {
            'bridge': NoOpBridgeConfig().to_schema(),
            'encoder': {
                'input_layers': ['image'],
                'output_layers': ['encoded'],
                'layers': [
                    {'Dense': {'units': 1, 'name': 'encoded'}}
                ]
            },
            'decoder': {
                'input_layers': ['image'],
                'output_layers': ['encoded'],
                'layers': [
                    {'Dense': {'units': 1, 'name': 'decoded'}}
                ]
            },
            'loss': MeanSquaredErrorConfig(input_layer=['image', 0, 0],
                                           output_layer=['decoded', 0, 0]).to_schema(),
            'optimizer': AdamConfig(learning_rate=0.01).to_schema(),
            'metrics': [],
            'summaries': ['loss', 'gradients'],
            'clip_gradients': 0.5,
            'clip_embed_gradients': 0.,
            'name': 'model'}
        config = GeneratorConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert config_dict.pop('bridge') == config_to_dict.pop('bridge')
        assert_equal_graphs(config_dict.pop('encoder'), config_to_dict.pop('encoder'))
        assert_equal_graphs(config_dict.pop('decoder'), config_to_dict.pop('decoder'))
        assert_equal_dict(config_dict, config_to_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import six

from marshmallow import ValidationError, fields, validate, validates_schema

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.exceptions import PolyaxonConfigurationError
from polyaxon_schemas.matrix import MatrixConfig
from polyaxon_schemas.polyaxonfile.utils import cached_property
from polyaxon_schemas.utils import (
    AcquisitionFunctions,
    EarlyStoppingPolicy,
    GaussianProcessesKernels,
    Optimization,
    ResourceTypes,
    SearchAlgorithms
)


class EarlyStoppingMetricSchema(BaseSchema):
    metric = fields.Str()
    value = fields.Float()
    optimization = fields.Str(allow_none=True, validate=validate.OneOf(Optimization.VALUES))
    policy = fields.Str(allow_none=True, validate=validate.OneOf(EarlyStoppingPolicy.VALUES))

    @staticmethod
    def schema_config():
        return EarlyStoppingMetricConfig


class EarlyStoppingMetricConfig(BaseConfig):
    """
    Early stopping metric config.

    Args:
        metric: `str`. The metric to use for early stopping.
        value: `float`. The metric value to use for the condition.
        optimization: `string`. The optimization to do: maximize or minimize.
    """
    SCHEMA = EarlyStoppingMetricSchema
    IDENTIFIER = 'early_stopping_metric'

    def __init__(self,
                 metric,
                 value=None,
                 optimization=Optimization.MAXIMIZE,
                 policy=EarlyStoppingPolicy.ALL):
        self.metric = metric
        self.value = value
        self.optimization = optimization
        self.policy = policy


class RandomSearchSchema(BaseSchema):
    n_experiments = fields.Int(allow_none=True, validate=validate.Range(min=1))

    @staticmethod
    def schema_config():
        return RandomSearchConfig


class RandomSearchConfig(BaseConfig):
    SCHEMA = RandomSearchSchema
    IDENTIFIER = 'random_search'
    REDUCED_ATTRIBUTES = ['n_experiments']

    def __init__(self, n_experiments=None):
        self.n_experiments = n_experiments


class GridSearchSchema(BaseSchema):
    n_experiments = fields.Int(allow_none=True, validate=validate.Range(min=1))

    @staticmethod
    def schema_config():
        return GridSearchConfig


class GridSearchConfig(BaseConfig):
    SCHEMA = GridSearchSchema
    IDENTIFIER = 'grid_search'
    REDUCED_ATTRIBUTES = ['n_experiments']

    def __init__(self, n_experiments=None):
        self.n_experiments = n_experiments


class SearchMetricSchema(BaseSchema):
    name = fields.Str()
    optimization = fields.Str(allow_none=True, validate=validate.OneOf(Optimization.VALUES))

    @staticmethod
    def schema_config():
        return SearchMetricConfig


class SearchMetricConfig(BaseConfig):
    SCHEMA = SearchMetricSchema
    IDENTIFIER = 'search_metric'

    def __init__(self,
                 name,
                 optimization=Optimization.MAXIMIZE):
        self.name = name
        self.optimization = optimization


class ResourceSchema(BaseSchema):
    name = fields.Str()
    type = fields.Str(allow_none=True, validate=validate.OneOf(ResourceTypes.VALUES))

    @staticmethod
    def schema_config():
        return ResourceConfig


class ResourceConfig(BaseConfig):
    SCHEMA = ResourceSchema
    IDENTIFIER = 'resource'

    def __init__(self,
                 name,
                 type=ResourceTypes.FLOAT):  # noqa, redefined-builtin `len`
        self.name = name
        self.type = type

    def cast_value(self, value):
        if ResourceTypes.is_int(self.type):
            return int(value)
        if ResourceTypes.is_float(self.type):
            return float(value)
        return value


class HyperbandSchema(BaseSchema):
    max_iter = fields.Int(validate=validate.Range(min=1))
    eta = fields.Float(validate=validate.Range(min=0))
    resource = fields.Nested(ResourceSchema)
    metric = fields.Nested(SearchMetricSchema)
    resume = fields.Boolean(allow_none=True)

    @staticmethod
    def schema_config():
        return HyperbandConfig


class HyperbandConfig(BaseConfig):
    SCHEMA = HyperbandSchema
    IDENTIFIER = 'hyperband'

    def __init__(self, max_iter, eta, resource, metric, resume=False):
        self.max_iter = max_iter
        self.eta = eta
        self.resource = resource
        self.metric = metric
        self.resume = resume


class GaussianProcessSchema(BaseSchema):
    kernel = fields.Str(allow_none=True, validate=validate.OneOf(GaussianProcessesKernels.VALUES))
    length_scale = fields.Float(allow_none=True)
    nu = fields.Float(allow_none=True)
    n_restarts_optimizer = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return GaussianProcessConfig


class GaussianProcessConfig(BaseConfig):
    SCHEMA = GaussianProcessSchema
    IDENTIFIER = 'gaussian_process'

    def __init__(self,
                 kernel=GaussianProcessesKernels.MATERN,
                 length_scale=1.0,
                 nu=1.5,
                 n_restarts_optimizer=0):
        self.kernel = kernel
        self.length_scale = length_scale
        self.nu = nu
        self.n_restarts_optimizer = n_restarts_optimizer


def validate_utility_function(acquisition_function, kappa, eps):
    condition = AcquisitionFunctions.is_ucb(acquisition_function) and kappa is None
    if condition:
        raise ValidationError('the acquisition function `ucb` requires a parameter `kappa`')

    condition = ((AcquisitionFunctions.is_ei(acquisition_function) or
                  AcquisitionFunctions.is_poi(acquisition_function)) and
                 eps is None)
    if condition:
        raise ValidationError('the acquisition function `{}` requires a parameter `eps`'.format(
            acquisition_function
        ))


class UtilityFunctionSchema(BaseSchema):
    acquisition_function = fields.Str(allow_none=True,
                                      validate=validate.OneOf(AcquisitionFunctions.VALUES))
    gaussian_process = fields.Nested(GaussianProcessSchema, allow_none=True)
    kappa = fields.Float(allow_none=True)
    eps = fields.Float(allow_none=True)
    n_warmup = fields.Int(allow_none=True)
    n_iter = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return UtilityFunctionConfig

    @validates_schema
    def validate_utility_function(self, data):
        validate_utility_function(
            acquisition_function=data.get('acquisition_function'),
            kappa=data.get('kappa'),
            eps=data.get('eps'))


class UtilityFunctionConfig(BaseConfig):
    SCHEMA = UtilityFunctionSchema
    IDENTIFIER = 'utility_function'
    REDUCED_ATTRIBUTES = ['n_warmup', 'n_iter']

    def __init__(self,
                 acquisition_function=AcquisitionFunctions.UCB,
                 gaussian_process=None,
                 kappa=None,
                 eps=None,
                 n_warmup=None,
                 n_iter=None):
        validate_utility_function(
            acquisition_function=acquisition_function,
            kappa=kappa,
            eps=eps)

        self.acquisition_function = acquisition_function
        self.gaussian_process = gaussian_process
        self.kappa = kappa
        self.eps = eps
        self.n_warmup = n_warmup
        self.n_iter = n_iter


class BOSchema(BaseSchema):
    utility_function = fields.Nested(UtilityFunctionSchema, allow_none=True)
    n_initial_trials = fields.Int()
    n_iterations = fields.Int()
    metric = fields.Nested(SearchMetricSchema)

    @staticmethod
    def schema_config():
        return BOConfig


class BOConfig(BaseConfig):
    SCHEMA = BOSchema
    IDENTIFIER = 'bo'

    def __init__(self, n_initial_trials, n_iterations, metric, utility_function=None):
        self.n_initial_trials = n_initial_trials
        self.n_iterations = n_iterations
        self.utility_function = utility_function
        self.metric = metric


def validate_search_algorithm(algorithms, matrix):
    used_algorithms = sum([1 for f in algorithms if f is not None])
    if used_algorithms > 1:
        raise ValidationError('Only one search algorithm can be used.')
    if used_algorithms and not matrix:
        raise ValidationError('Search algorithms need a matrix definition.')


def validate_bo_matrix(matrix):
    if not matrix:
        return None

    matrix_data = {}
    for key, value in six.iteritems(matrix):
        if not isinstance(value, MatrixConfig):
            matrix_data[key] = MatrixConfig.from_dict(value)
        else:
            matrix_data[key] = value

        if matrix_data[key].is_distribution and not matrix_data[key].is_uniform:
            raise ValidationError('`{}` defines a non uniform distribution, '
                                  'and it cannot be used with bayesian optimization.'.format(key))

    return matrix_data


def validate_matrix(matrix, is_grid_search=False, is_bo=False):
    if not matrix:
        return None

    matrix_data = {}
    for key, value in six.iteritems(matrix):
        if not isinstance(value, MatrixConfig):
            matrix_data[key] = MatrixConfig.from_dict(value)
        else:
            matrix_data[key] = value

        if is_grid_search and matrix_data[key].is_distribution:
            raise ValidationError('`{}` defines a distribution, '
                                  'and it cannot be used with grid search.'.format(key))

        if is_bo and matrix_data[key].is_distribution and not matrix_data[key].is_uniform:
            raise ValidationError('`{}` defines a non uniform distribution, '
                                  'and it cannot be used with bayesian optimization.'.format(key))

    return matrix_data


class HPTuningSchema(BaseSchema):
    seed = fields.Int(allow_none=True)
    matrix = fields.Dict(allow_none=True)
    concurrency = fields.Int(allow_none=True)
    grid_search = fields.Nested(GridSearchSchema, allow_none=None)
    random_search = fields.Nested(RandomSearchSchema, allow_none=None)
    hyperband = fields.Nested(HyperbandSchema, allow_none=None)
    bo = fields.Nested(BOSchema, allow_none=None)
    early_stopping = fields.Nested(EarlyStoppingMetricSchema, many=True, allow_none=True)

    @staticmethod
    def schema_config():
        return HPTuningConfig

    @validates_schema
    def validate_search_algorithm(self, data):
        validate_search_algorithm(
            algorithms=[data.get('grid_search'),
                        data.get('random_search'),
                        data.get('hyperband'),
                        data.get('bo')],
            matrix=data.get('matrix'))

    @validates_schema
    def validate_matrix(self, data):
        """Validates matrix data and creates the config objects"""
        is_grid_search = data.get('grid_search') is not None
        is_bo = data.get('bo') is not None
        validate_matrix(data.get('matrix'), is_grid_search=is_grid_search, is_bo=is_bo)


class HPTuningConfig(BaseConfig):
    SCHEMA = HPTuningSchema
    IDENTIFIER = 'hptuning'
    REDUCED_ATTRIBUTES = ['grid_search', 'random_search', 'hyperband', 'bo']

    def __init__(self,
                 seed=None,
                 matrix=None,
                 concurrency=1,
                 grid_search=None,
                 random_search=None,
                 hyperband=None,
                 bo=None,
                 early_stopping=None):
        self.seed = seed
        matrix = validate_matrix(matrix,
                                 is_grid_search=grid_search is not None,
                                 is_bo=bo is not None)
        self.matrix = matrix
        self.concurrency = concurrency
        validate_search_algorithm(
            algorithms=[grid_search, random_search, hyperband, bo],
            matrix=matrix)
        self.grid_search = grid_search
        self.random_search = random_search
        self.hyperband = hyperband
        self.bo = bo
        self.early_stopping = early_stopping

    def to_dict(self, humanize_values=False):
        results = super(HPTuningConfig, self).to_dict(humanize_values=humanize_values)
        if not results.get('matrix'):
            return results
        results['matrix'] = {k: v.to_dict() for k, v in six.iteritems(results['matrix'])}
        return results

    @cached_property
    def search_algorithm(self):
        if not self.matrix:
            raise PolyaxonConfigurationError('a search algorithm requires a matrix definition.')
        if self.random_search:
            return SearchAlgorithms.RANDOM
        if self.hyperband:
            return SearchAlgorithms.HYPERBAND
        if self.grid_search:
            return SearchAlgorithms.GRID
        if self.bo:
            return SearchAlgorithms.BO
        # Default value
        return SearchAlgorithms.GRID



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import base64

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema


class LogHandlerSchema(BaseSchema):
    dsn = fields.Str(allow_none=True)
    environment = fields.Str()
    tags = fields.Dict()

    @staticmethod
    def schema_config():
        return LogHandlerConfig


class LogHandlerConfig(BaseConfig):
    SCHEMA = LogHandlerSchema
    IDENTIFIER = 'log_handler'

    def __init__(self, environment, tags, dsn=None):
        self.dsn = dsn
        self.environment = environment
        self.tags = tags

    @property
    def decoded_dsn(self):
        if self.dsn:
            return base64.b64decode(self.dsn.encode('utf-8')).decode('utf-8')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.base import BaseConfig, BaseSchema


class LoggingSchema(BaseSchema):
    level = fields.Str(allow_none=True, validate=validate.OneOf(
        ['INFO', 'DEBUG', 'WARN', 'ERROR', 'FATAL']))
    formatter = fields.Str(allow_none=True)

    @staticmethod
    def schema_config():
        return LoggingConfig


class LoggingConfig(BaseConfig):
    SCHEMA = LoggingSchema
    IDENTIFIER = 'logging'

    def __init__(self, level='INFO', formatter=None):
        self.level = level
        self.formatter = formatter



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import ValidationError, fields, validate, validates_schema

from polyaxon_schemas.base import BaseConfig, BaseSchema


def validate_image(image):
    if not image:
        raise ValidationError('Invalid docker image `{}`'.format(image))
    tagged_image = image.split(':')
    if len(tagged_image) > 3:
        raise ValidationError('Invalid docker image `{}`'.format(image))
    if len(tagged_image) == 3 and ('/' not in tagged_image[1] or tagged_image[1].startswith('/')):
        raise ValidationError('Invalid docker image `{}`'.format(image))


class BuildSchema(BaseSchema):
    image = fields.Str()
    build_steps = fields.List(fields.Str(), allow_none=True)
    env_vars = fields.List(fields.List(fields.Raw(), validate=validate.Length(equal=2)),
                           allow_none=True)
    git = fields.Str(allow_none=True)
    ref = fields.Str(allow_none=True)
    nocache = fields.Boolean(allow_none=True)

    @staticmethod
    def schema_config():
        return BuildConfig

    @validates_schema
    def validate_image(self, data):
        """Validates docker image structure"""
        validate_image(data.get('image'))


class BuildConfig(BaseConfig):
    """
    Build config.

    Args:
        image: str. The name of the image to use during the build step.
        build_steps: list(str). The build steps to apply to your docker image.
            (translate to multiple RUN ...)
        env_vars: list((str, str)) The environment variable to set on you docker image.
        nocache: `bool`. To not use cache when building the image.
        git: `str`. A url to git code, in case you are not using upload.
        ref: `str`. The commit/branch/treeish to use.

    """
    SCHEMA = BuildSchema
    IDENTIFIER = 'build'
    REDUCED_ATTRIBUTES = ['build_steps', 'env_vars', 'git', 'nocache', 'ref']

    def __init__(self, image, build_steps=None, env_vars=None, nocache=None, git=None, ref=None):
        validate_image(image)
        self.image = image
        self.build_steps = build_steps
        self.env_vars = env_vars
        self.git = git
        self.nocache = nocache
        self.ref = ref

    @property
    def image_tag(self):
        tagged_image = self.image.split(':')
        if len(tagged_image) == 1:
            return 'latest'
        if len(tagged_image) == 2:
            return 'latest' if '/' in tagged_image[-1] else tagged_image[-1]
        if len(tagged_image) == 3:
            return tagged_image[-1]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import itertools

from polyaxon_schemas.base import BaseConfig, BaseSchema


class ForSchema(BaseSchema):
    class Meta:
        ordered = True
        fields = ('len', 'do', 'index')

    @staticmethod
    def schema_config():
        return ForConfig


class ForConfig(BaseConfig):
    SCHEMA = ForSchema
    IDENTIFIER = 'for'

    def __init__(self, len, do, index='index'):  # noqa, redefined-builtin `len`
        self.len = len
        self.do = do
        self.index = index

    def parse(self, spec, parser, declarations):
        parsed_data = []
        length = parser.parse_expression(spec, self.len, declarations, check_operators=True)
        length = int(length)
        for i in range(length):
            i_declarations = {self.index: i}
            i_declarations.update(declarations)
            parsed_data.append(
                parser.parse_expression(spec, self.do, i_declarations, check_operators=True))
        if parsed_data and isinstance(parsed_data[0], (list, tuple)):
            return list(itertools.chain.from_iterable(parsed_data))
        return parsed_data


class IfSchema(BaseSchema):
    class Meta:
        ordered = True
        fields = ('cond', 'do', 'else_do')

    @staticmethod
    def schema_config():
        return IfConfig


class IfConfig(BaseConfig):
    SCHEMA = IfSchema
    IDENTIFIER = 'if'

    def __init__(self, cond, do, else_do=None):
        self.cond = cond
        self.do = do
        self.else_do = else_do

    @staticmethod
    def _check_cond(spec, parser, cond):
        cond_template = "{{% if {} %}}1{{% else %}}0{{% endif %}}"
        cond_result = parser.parse_expression(spec, cond_template.format(cond), {})
        if int(cond_result) == 1:
            return True
        return False

    def parse(self, spec, parser, declarations):
        cond = parser.parse_expression(spec, self.cond, declarations, check_operators=False)
        if self._check_cond(spec, parser, cond):
            return parser.parse_expression(spec, self.do, declarations, check_operators=True)
        if self.else_do:
            return parser.parse_expression(spec, self.else_do, declarations, check_operators=True)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import Schema, fields, post_dump, post_load

from polyaxon_schemas.base import BaseConfig, BaseMultiSchema


class BaseOptimizerSchema(Schema):
    learning_rate = fields.Float(allow_none=True)
    decay_type = fields.Str(allow_none=True)
    decay_rate = fields.Float(allow_none=True)
    decay_steps = fields.Int(allow_none=True)
    start_decay_at = fields.Int(allow_none=True)
    stop_decay_at = fields.Int(allow_none=True)
    min_learning_rate = fields.Float(allow_none=True)
    staircase = fields.Bool(allow_none=True)
    global_step = fields.Str(allow_none=True)
    use_locking = fields.Bool(allow_none=True)
    name = fields.Str(allow_none=True)


class BaseOptimizerConfig(BaseConfig):
    REDUCED_ATTRIBUTES = ['name']

    def __init__(self,
                 learning_rate=0.001,
                 decay_type="",
                 decay_rate=0.,
                 decay_steps=100,
                 start_decay_at=0,
                 stop_decay_at=1e10,
                 min_learning_rate=1e-12,
                 staircase=False,
                 global_step=None,
                 use_locking=False,
                 name='optimizer'):
        self.learning_rate = learning_rate
        self.decay_type = decay_type
        self.decay_rate = decay_rate
        self.decay_steps = decay_steps
        self.start_decay_at = start_decay_at
        self.stop_decay_at = stop_decay_at
        self.min_learning_rate = min_learning_rate
        self.staircase = staircase
        self.global_step = global_step
        self.use_locking = use_locking
        self.name = name


class SGDSchema(BaseOptimizerSchema):
    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return SGDConfig(**data)

    @post_dump
    def unmake(self, data):
        return SGDConfig.remove_reduced_attrs(data)


class SGDConfig(BaseOptimizerConfig):
    """Optimizer that implements the gradient descent algorithm.

    Args:
        learning_rate: A Tensor or a floating point value. The learning rate to use.
        decay_type: A decay function name defined in `tf.train`
        decay_rate: `float`. The learning rate decay to apply.
        decay_steps: `int`. Apply decay every provided steps.
        start_decay_at: `int`. Don't decay before this step.
        stop_decay_at: `int`. Don't decay after this step.
        min_learning_rate: `float`. Don't decay below this number.
        staircase: `bool`. It `True` decay learning rate at discrete intervals.
        global_step: Scalar int `Tensor`, step counter for each update.
        use_locking: If True use locks for update operations.
        name: Optional name prefix for the operations created when applying gradients.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      optimizer: SGD
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        SGD:
          learning_rate: 0.21
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        SGD: {learning_rate: 0.21}
      # other model properties
    ```
    """
    IDENTIFIER = 'SGD'
    SCHEMA = SGDSchema

    def __init__(self,  # pylint: disable=useless-super-delegation
                 learning_rate=0.01,
                 decay_type="",
                 decay_rate=0.,
                 decay_steps=100,
                 start_decay_at=0,
                 stop_decay_at=1e10,
                 min_learning_rate=1e-12,
                 staircase=False,
                 global_step=None,
                 use_locking=False,
                 name='SGD'):
        super(SGDConfig, self).__init__(learning_rate, decay_type, decay_rate, decay_steps,
                                        start_decay_at, stop_decay_at, min_learning_rate,
                                        staircase, global_step, use_locking, name)


class MomentumSchema(BaseOptimizerSchema):
    momentum = fields.Float(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return MomentumConfig(**data)

    @post_dump
    def unmake(self, data):
        return MomentumConfig.remove_reduced_attrs(data)


class MomentumConfig(BaseOptimizerConfig):
    """Optimizer that implements the Momentum.

    Momentum Optimizer accepts learning rate decay. When training a model,
    it is often recommended to lower the learning rate as the training
    progresses. The function returns the decayed learning rate.  It is
    computed as:

    ```python
    >>> decayed_learning_rate = learning_rate * decay_rate ^ (global_step / lr_decay_steps)
    ```

    Args:
        learning_rate: `float`. Learning rate.
        momentum: `float`. Momentum.
        decay_type: A decay function name defined in `tf.train`
        decay_rate: `float`. The learning rate decay to apply.
        decay_steps: `int`. Apply decay every provided steps.
        start_decay_at: `int`. Don't decay before this step.
        stop_decay_at: `int`. Don't decay after this step.
        min_learning_rate: `float`. Don't decay below this number.
        staircase: `bool`. It `True` decay learning rate at discrete intervals.
        global_step: Scalar int `Tensor`, step counter for each update.
        use_locking: If True use locks for update operations.
        name: `str`. Optional name prefix for the operations created when applying gradients.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      optimizer: Momentum
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        Momentum:
          learning_rate: 0.21
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        Momentum: {learning_rate: 0.21}
      # other model properties
    ```
    """
    IDENTIFIER = 'Momentum'
    SCHEMA = MomentumSchema

    def __init__(self,
                 learning_rate=0.001,
                 momentum=0.9,
                 decay_type="",
                 decay_rate=0.,
                 decay_steps=10000,
                 start_decay_at=0,
                 stop_decay_at=1e10,
                 min_learning_rate=1e-12,
                 staircase=False,
                 global_step=None,
                 use_locking=False,
                 name='Momentum'):
        self.momentum = momentum
        super(MomentumConfig, self).__init__(learning_rate, decay_type, decay_rate, decay_steps,
                                             start_decay_at, stop_decay_at, min_learning_rate,
                                             staircase, global_step, use_locking, name)


class NestrovSchema(BaseOptimizerSchema):
    momentum = fields.Float(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return NestrovConfig(**data)

    @post_dump
    def unmake(self, data):
        return NestrovConfig.remove_reduced_attrs(data)


class NestrovConfig(BaseOptimizerConfig):
    """Optimizer that implements the Nesterov.

    Same as Momentum optimizer but uses nestrov
    See [Sutskever et. al., 2013](http://jmlr.org/proceedings/papers/v28/sutskever13.pdf)

    ```python
    >>> decayed_learning_rate = learning_rate * decay_rate ^ (global_step / lr_decay_steps)
    ```
    Args:
        learning_rate: `float`. Learning rate.
        momentum: `float`. Momentum.
        decay_type: A decay function name defined in `tf.train`
        decay_rate: `float`. The learning rate decay to apply.
        decay_steps: `int`. Apply decay every provided steps.
        start_decay_at: `int`. Don't decay before this step.
        stop_decay_at: `int`. Don't decay after this step.
        min_learning_rate: `float`. Don't decay below this number.
        staircase: `bool`. It `True` decay learning rate at discrete intervals.
        global_step: Scalar int `Tensor`, step counter for each update.
        use_locking: If True use locks for update operations.
        name: `str`. Optional name prefix for the operations created when applying gradients.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      optimizer: Nestrov
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        Nestrov:
          learning_rate: 0.21
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        Nestrov: {learning_rate: 0.21}
      # other model properties
    ```
    """
    IDENTIFIER = 'Nestrov'
    SCHEMA = NestrovSchema

    def __init__(self,
                 learning_rate=0.001,
                 momentum=0.9,
                 decay_type="",
                 decay_rate=0.,
                 decay_steps=10000,
                 start_decay_at=0,
                 stop_decay_at=1e10,
                 min_learning_rate=1e-12,
                 staircase=False,
                 use_locking=False,
                 global_step=None,
                 name='Nestrov'):
        self.momentum = momentum
        super(NestrovConfig, self).__init__(learning_rate, decay_type, decay_rate, decay_steps,
                                            start_decay_at, stop_decay_at, min_learning_rate,
                                            staircase, global_step, use_locking, name)


class RMSPropSchema(BaseOptimizerSchema):
    decay = fields.Float(allow_none=True)
    momentum = fields.Float(allow_none=True)
    epsilon = fields.Float(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return RMSPropConfig(**data)

    @post_dump
    def unmake(self, data):
        return RMSPropConfig.remove_reduced_attrs(data)


class RMSPropConfig(BaseOptimizerConfig):
    """Optimizer that implements the RMSprop.

    Maintain a moving (discounted) average of the square of gradients.
    Divide gradient by the root of this average.

    Args:
        learning_rate: `float`. learning rate.
        decay: `float`. Discounting factor for the history/coming gradient.
        momentum: `float`. Momentum.
        epsilon: `float`. Small value to avoid zero denominator.
        decay_type: A decay function name defined in `tf.train`
        decay_rate: `float`. The learning rate decay to apply.
        decay_steps: `int`. Apply decay every provided steps.
        start_decay_at: `int`. Don't decay before this step.
        stop_decay_at: `int`. Don't decay after this step.
        min_learning_rate: `float`. Don't decay below this number.
        staircase: `bool`. It `True` decay learning rate at discrete intervals.
        global_step: Scalar int `Tensor`, step counter for each update.
        use_locking: If True use locks for update operations.
        name: Optional name prefix for the operations created when applying gradients.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      optimizer: RMSProp
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        RMSProp:
          learning_rate: 0.21
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        RMSProp: {learning_rate: 0.21}
      # other model properties
    ```
    """
    IDENTIFIER = 'RMSProp'
    SCHEMA = RMSPropSchema

    def __init__(self,
                 learning_rate=0.001,
                 decay=0.9,
                 momentum=0.0,
                 epsilon=1e-10,
                 decay_type="",
                 decay_rate=0.,
                 decay_steps=10000,
                 start_decay_at=0,
                 stop_decay_at=1e10,
                 min_learning_rate=1e-12,
                 staircase=False,
                 global_step=None,
                 use_locking=False,
                 name='RMSProp'):
        self.decay = decay
        self.momentum = momentum
        self.epsilon = epsilon
        super(RMSPropConfig, self).__init__(learning_rate, decay_type, decay_rate, decay_steps,
                                            start_decay_at, stop_decay_at, min_learning_rate,
                                            staircase, global_step, use_locking, name)


class AdamSchema(BaseOptimizerSchema):
    beta1 = fields.Float(allow_none=True)
    beta2 = fields.Float(allow_none=True)
    epsilon = fields.Float(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return AdamConfig(**data)

    @post_dump
    def unmake(self, data):
        return AdamConfig.remove_reduced_attrs(data)


class AdamConfig(BaseOptimizerConfig):
    """Optimizer that implements the Adam.

    The default value of 1e-8 for epsilon might not be a good default in
    general. For example, when training an Inception network on ImageNet a
    current good choice is 1.0 or 0.1.

    Args:
        learning_rate: `float`. learning rate.
        beta1: `float`. The exponential decay rate for the 1st moment estimates.
        beta2: `float`. The exponential decay rate for the 2nd moment estimates.
        epsilon: `float`. A small constant for numerical stability.
        epsilon: `float`. Small value to avoid zero denominator.
        decay_type: A decay function name defined in `tf.train`
        decay_rate: `float`. The learning rate decay to apply.
        decay_steps: `int`. Apply decay every provided steps.
        start_decay_at: `int`. Don't decay before this step.
        stop_decay_at: `int`. Don't decay after this step.
        min_learning_rate: `float`. Don't decay below this number.
        staircase: `bool`. It `True` decay learning rate at discrete intervals.
        global_step: Scalar int `Tensor`, step counter for each update.
        use_locking: If True use locks for update operations.
        name: `str`. Optional name prefix for the operations created when applying gradients.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      optimizer: Adam
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        Adam:
          learning_rate: 0.21
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        Adam: {learning_rate: 0.21}
      # other model properties
    ```
    """
    IDENTIFIER = 'Adam'
    SCHEMA = AdamSchema

    def __init__(self,
                 learning_rate=0.001,
                 beta1=0.9,
                 beta2=0.999,
                 epsilon=1e-8,
                 decay_type="",
                 decay_rate=0.,
                 decay_steps=10000,
                 start_decay_at=0,
                 stop_decay_at=1e10,
                 min_learning_rate=1e-12,
                 staircase=False,
                 global_step=None,
                 use_locking=False,
                 name='Adam'):
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        super(AdamConfig, self).__init__(learning_rate, decay_type, decay_rate, decay_steps,
                                         start_decay_at, stop_decay_at, min_learning_rate,
                                         staircase, global_step, use_locking, name)


class AdagradSchema(BaseOptimizerSchema):
    initial_accumulator_value = fields.Float(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return AdagradConfig(**data)

    @post_dump
    def unmake(self, data):
        return AdagradConfig.remove_reduced_attrs(data)


class AdagradConfig(BaseOptimizerConfig):
    """Optimizer that implements AdaGrad.

    Args:
        learning_rate: `float`. Learning rate.
        initial_accumulator_value: `float`. Starting value for the
            accumulators, must be positive.
        decay_type: A decay function name defined in `tf.train`
        decay_rate: `float`. The learning rate decay to apply.
        decay_steps: `int`. Apply decay every provided steps.
        start_decay_at: `int`. Don't decay before this step.
        stop_decay_at: `int`. Don't decay after this step.
        min_learning_rate: `float`. Don't decay below this number.
        staircase: `bool`. It `True` decay learning rate at discrete intervals.
        global_step: Scalar int `Tensor`, step counter for each update.
        use_locking: If True use locks for update operations.
        name: `str`. Optional name prefix for the operations created when applying gradients.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      optimizer: Adagrad
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        Adagrad:
          learning_rate: 0.21
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        Adagrad: {learning_rate: 0.21}
      # other model properties
    ```
    """
    IDENTIFIER = 'Adagrad'
    SCHEMA = AdagradSchema

    def __init__(self,
                 learning_rate=0.01,
                 initial_accumulator_value=0.1,
                 decay_type="",
                 decay_rate=0.,
                 decay_steps=10000,
                 start_decay_at=0,
                 stop_decay_at=1e10,
                 min_learning_rate=1e-12,
                 staircase=False,
                 global_step=None,
                 use_locking=False,
                 name='Adagrad'):
        self.initial_accumulator_value = initial_accumulator_value
        super(AdagradConfig, self).__init__(learning_rate, decay_type, decay_rate, decay_steps,
                                            start_decay_at, stop_decay_at, min_learning_rate,
                                            staircase, global_step, use_locking, name)


class AdadeltaSchema(BaseOptimizerSchema):
    rho = fields.Float(allow_none=True)
    epsilon = fields.Float(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return AdadeltaConfig(**data)

    @post_dump
    def unmake(self, data):
        return AdadeltaConfig.remove_reduced_attrs(data)


class AdadeltaConfig(BaseOptimizerConfig):
    """Optimizer that implements Ftrl Proximal.

    The Ftrl-proximal algorithm, abbreviated for Follow-the-regularized-leader,
    is described in the paper below.

    It can give a good performance vs. sparsity tradeoff.

    Ftrl-proximal uses its own global base learning rate and can behave like
    Adagrad with `learning_rate_power=-0.5`, or like gradient descent with
    `learning_rate_power=0.0`.

    Args:
        learning_rate: `float`. Learning rate.
        learning_rate_power: `float`. Must be less or equal to zero.
        initial_accumulator_value: `float`. The starting value for accumulators.
            Only positive values are allowed.
        l1_regularization_strength: `float`. Must be less or equal to zero.
        l2_regularization_strength: `float`. Must be less or equal to zero.
        decay_type: A decay function name defined in `tf.train`
        decay_rate: `float`. The learning rate decay to apply.
        decay_steps: `int`. Apply decay every provided steps.
        start_decay_at: `int`. Don't decay before this step.
        stop_decay_at: `int`. Don't decay after this step.
        min_learning_rate: `float`. Don't decay below this number.
        staircase: `bool`. It `True` decay learning rate at discrete intervals.
        global_step: Scalar int `Tensor`, step counter for each update.
        use_locking: If True use locks for update operations.
        name: `str`. Optional name prefix for the operations created when applying gradients.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      optimizer: Adadelta
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        Adadelta:
          learning_rate: 0.21
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        Adadelta: {learning_rate: 0.21}
      # other model properties
    ```
    """
    IDENTIFIER = 'Adadelta'
    SCHEMA = AdadeltaSchema

    def __init__(self,
                 learning_rate=0.99,
                 rho=0.95,
                 epsilon=1e-08,
                 decay_type="",
                 decay_rate=0.,
                 decay_steps=10000,
                 start_decay_at=0,
                 stop_decay_at=1e10,
                 min_learning_rate=1e-12,
                 staircase=False,
                 global_step=None,
                 use_locking=False,
                 name='Adadelta'):
        self.rho = rho
        self.epsilon = epsilon
        super(AdadeltaConfig, self).__init__(learning_rate, decay_type, decay_rate, decay_steps,
                                             start_decay_at, stop_decay_at, min_learning_rate,
                                             staircase, global_step, use_locking, name)


class FtrlSchema(BaseOptimizerSchema):
    learning_rate_power = fields.Float(allow_none=True)
    initial_accumulator_value = fields.Float(allow_none=True)
    l1_regularization_strength = fields.Float(allow_none=True)
    l2_regularization_strength = fields.Float(allow_none=True)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return FtrlConfig(**data)

    @post_dump
    def unmake(self, data):
        return FtrlConfig.remove_reduced_attrs(data)


class FtrlConfig(BaseOptimizerConfig):
    """Optimizer that implements AdaDelta.

    Args:
        learning_rate: A `Tensor` or a floating point value. The learning rate.
        rho: A `Tensor` or a floating point value. The decay rate.
        epsilon: A `Tensor` or a floating point value.  A constant epsilon used to better
            conditioning the grad update.
        decay_type: A decay function name defined in `tf.train`
        decay_rate: `float`. The learning rate decay to apply.
        decay_steps: `int`. Apply decay every provided steps.
        start_decay_at: `int`. Don't decay before this step.
        stop_decay_at: `int`. Don't decay after this step.
        min_learning_rate: `float`. Don't decay below this number.
        staircase: `bool`. It `True` decay learning rate at discrete intervals.
        global_step: Scalar int `Tensor`, step counter for each update.
        use_locking: If True use locks for update operations.
        name: Optional name prefix for the operations created when applying gradients.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      optimizer: Ftrl
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        Ftrl:
          learning_rate: 0.21
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      optimizer:
        Ftrl: {learning_rate: 0.21}
      # other model properties
    ```
    """
    IDENTIFIER = 'Ftrl'
    SCHEMA = FtrlSchema

    def __init__(self,
                 learning_rate=3.0,
                 learning_rate_power=-0.5,
                 initial_accumulator_value=0.1,
                 l1_regularization_strength=0.0,
                 l2_regularization_strength=0.0,
                 decay_type="",
                 decay_rate=0.,
                 decay_steps=10000,
                 start_decay_at=0,
                 stop_decay_at=1e10,
                 min_learning_rate=1e-12,
                 staircase=False,
                 global_step=None,
                 use_locking=False,
                 name='Ftrl'):
        self.learning_rate_power = learning_rate_power
        self.initial_accumulator_value = initial_accumulator_value
        self.l1_regularization_strength = l1_regularization_strength
        self.l2_regularization_strength = l2_regularization_strength
        super(FtrlConfig, self).__init__(learning_rate, decay_type, decay_rate, decay_steps,
                                         start_decay_at, stop_decay_at, min_learning_rate,
                                         staircase, global_step, use_locking, name)


class OptimizerSchema(BaseMultiSchema):
    __multi_schema_name__ = 'optimizer'
    __configs__ = {
        SGDConfig.IDENTIFIER: SGDConfig,
        MomentumConfig.IDENTIFIER: MomentumConfig,
        NestrovConfig.IDENTIFIER: NestrovConfig,
        RMSPropConfig.IDENTIFIER: RMSPropConfig,
        AdamConfig.IDENTIFIER: AdamConfig,
        AdagradConfig.IDENTIFIER: AdagradConfig,
        AdadeltaConfig.IDENTIFIER: AdadeltaConfig,
        FtrlConfig.IDENTIFIER: FtrlConfig,
    }
